from ..savable import Savable
from pathlib import Path

class FileStorage(Savable):

    path: str

    def __init__(self, path:str):
        self.path = path
    
    def exists(self, path: str) -> bool:
        return Path(path).exists()


from __future__ import annotations
from abc import abstractmethod
from typing import Dict, Any, Type, cast

from epure.errors import DbError
from .table_column import TableColumn

from ...named import FullName, SnakeCaseNamed
from ..savable import Savable
from .table import Table
from .db_entity_resource import DbEntityResource
from ...helpers.type_helper import check_subclass, check_type
from ...epure import Epure

class TableStorage(DbEntityResource):

    tables:Dict[str,Table]
    namespaces:list[str]
    default_table_type:Type[Table] = Table
    migrate_on_delete:bool = False
    _epures: Dict[str, type] = None

    @abstractmethod
    def __init__(self, default_table_type:Type[Table]=None, 
    migrate_on_delete:bool=False):

        if default_table_type:
            self.default_table_type = default_table_type

        self.migrate_on_delete = migrate_on_delete

        super().__init__() 


        
    def __getitem__(self, table_name:str):
        full_name = self._get_full_table_name(table_name)
        return self.tables[full_name.full_name]

    def __iter__(self):
        return self.tables.__iter__()

    def __len__(self):
        return self.tables.__len__()

    def __contains__(self, table_name:str) -> bool:
        full_name = self._get_full_table_name(table_name)
        if full_name.full_name in self.tables:
            return True
        table = self.read(table_name)
        return bool(table)
    
    def connect(self) -> None:
        Epure.EDb = self
        
    def create_namespace(self, namespace):
        script = self.serialize_namespace_for_create(namespace)
        script = str(script)
        self.execute(script)

        self._set_namespace(namespace)

    def create_table(self, table: Savable):
        check_type('table', table, self.default_table_type)
        table = cast(Table, table)

        script = self.serialize_for_create(table)
        script = str(script)
        self.execute(script)

        self._set_table(table)
        return table

    def update_table(self, py_table: Savable) -> Any:
        check_type('py_table', py_table, Table)

        py_table = cast(Table, py_table)        
        db_table = self.tables[py_table.full_name]

        py_header = py_table.header
        db_header = db_table.header

        diff = py_header - db_header

        for column in diff:
            if column.name in db_header:
                db_header.update(column)
            else:
                db_header.create(column)
        
        diff = db_header - py_header
        for column in diff:
            db_header.delete(column)

        self._set_table(py_table)
        return py_table


    def deserialize_table(self, table_columns: object) -> Table:
        check_type('table_columns', table_columns, list)
        table_columns = cast(list, table_columns)

        full_name = self._deserialize_table_name(table_columns)

        TableCls = self._get_table_type(table_columns)
        table = TableCls(name=full_name.name, namespace=full_name.namespace, resource=self)
        
        
        for column_dict in table_columns:
            column = table.header.deserialize(column_dict, db=self)
            column = cast(TableColumn, column)
            table.header._set_column(column)
        
        return table


    def get_epure_by_table_name(self, table_name: str):
        if self._epures == None:            
            self._epures = self._get_epures_dict()
        if table_name not in self._epures:
            from ...epure import Epure
            self._epures[table_name] = self._get_epure_by_table_name()
        if table_name not in self._epures:
            return None
            # raise DbError('no one epure has this table as resource')
        return self._epures[table_name]
    
    def _get_epures_dict(self):        
        res = {}
        for ep in Epure.epures:
            if not hasattr(ep, 'resource'):
                raise DbError(f'epure {ep} not correctly initialized')
            table_name = ep.resource.full_name
            if table_name in self:
                res[table_name] = ep
        # if not res:
        #     raise DbError('no one epure has table as resource')
        return res
    
    def _get_epure_by_table_name(self, table_name: str):
        for ep in Epure.epures:
            if ep.resource.full_name in self:
                return ep
        return None
        
    def _get_table_type(self, table_columns:list):
        return self.default_table_type

    def _deserialize_table_name(self, table_columns:list) -> FullName:
        raise NotImplementedError
    
    def _set_namespace(self, namespace):
        self.namespaces.append(namespace)

    def _set_table(self, table:Table):
        self.tables[table.full_name] = table
        if table.namespace == self.get_default_namespace():
            self.tables[table.name] = table
        table.resource = self


    # def get_table_for_resource(self, resource: Savable, table_name: str = '') -> Table:
    #     table:Table

    #     full_name = self._get_full_table_name(table_name)
    #     columns = self.get_annotations(resource)
    #     columns = {name: val
    #         for name, val in columns.items() if
    #         not resource.is_excluded(name)}

    #     TableCls = self.default_table_type
    #     table = TableCls(full_name.name, columns, full_name.namespace)
    #     return table

    # def get_annotations(self, resource: Savable):
        
    #     res = {}
    #     resource_cls:type
    #     if isinstance(resource, type):
    #         resource_cls = resource
    #     resource_cls = resource.__class__

    #     res = resource.__annotations__
        
    #     for cls in resource_cls.mro():
    #         res.update(cls.__annotations__)


    def _get_full_table_name(self, table_name:str) -> FullName:
        # table_name = table_name
        full_name = table_name.split('.')
        if len(full_name) > 2:
            raise NameError('table_name name must have no more then one dot')
        if len(full_name) == 1:
            default_namespace = self.get_default_namespace()
            full_name.insert(0, default_namespace)

        res = SnakeCaseNamed(namespace=full_name[0], name=full_name[1])
        return res


from inflection import underscore
from abc import abstractmethod

class Named:
    name:str
    namespace:str

    @abstractmethod
    def __init__(self, name:str=None, namespace:str=None) -> None:
        if name is not None:
            self.name = name
        if namespace is not None:
            self.namespace = namespace

    @property
    def full_name(self):
        res = None
        if hasattr(self, 'name'):
            res = self.name
        if hasattr(self, 'namespace') and self.namespace and res:
            res = self.namespace + '.' + res

        return res
    
class NamedByClass(Named):

    @property
    def full_name(self):
        res = super().full_name
        if res is not None:
            return res
        if isinstance(self, type):
            return self.__name__
        else:
            return self.__class__.__name__

    
    
class FullName(Named):
    pass

class SnakeCaseNamed(FullName):
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = underscore(val)

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, val):
        self._namespace = underscore(val)


from .gres_table import GresTable
from  .gres_header import GresHeader
from typing import Any, Union, Dict, List
from types import NoneType
from ...helpers.type_helper import check_type
from ..db.table_header import TableHeader
from ..db.table_column import TableColumn
from ..db.table_storage import TableStorage
from ..resource import Resource

class JsonbHeader(GresHeader):
    def __contains__(self, other: Any) -> bool:
        if isinstance(other, str) or isinstance(other, TableColumn):
            return True
        raise NotImplementedError

    def __getitem__(self, key:str):
        res = TableColumn(key, str)
        self.columns[key] = res
        return self.columns[key]


class JsonbTable(GresTable):
    header: JsonbHeader

    def __init__(self, name: str,
            header:Union[TableHeader, Dict[str, Any]]=None, resource:Resource=None, namespace:str = '') -> None:        
        check_type('header', header, [TableHeader, dict, NoneType])        
        # header['jsonb___data'] = Any
        header = JsonbHeader(table=self)

        super().__init__(name=name, header=header, namespace=namespace, resource=resource)
    
    def serialize_header(self, db: TableStorage=None, **kwargs) -> List[Dict[str, str]]:
        res: List[Dict[str, str]] = [{'column_name': 'node_id', 'column_type': 'uuid'}, {'column_name': 'jsonb___data', 'column_type': 'jsonb'}]
        
        return res

        


from __future__ import annotations
from types import LambdaType, NoneType
from typing import TYPE_CHECKING, Dict, Union, List, ItemsView, Any, Type, Callable, cast
from ..savable import Savable
from .constraint import Constraint
from ..resource import Resource
from ..node.node import Node
from ...parser.term import Term
from ...parser.leaf import TableProxy, QueryingProxy, DbProxy, ColumnProxy
from ..db.table_column import TableColumn
from collections import OrderedDict
from ..node.proto import Proto

from ...errors import DeserializeError

if TYPE_CHECKING:
    from .table_storage import TableStorage
    from ...parser.term_parser import TermParser
from .db_entity import DbEntity
from ...helpers.type_helper import check_type
from ...errors import EpureError, DbError
from .table_header import TableHeader
from ..node_promise import FieldPromise, NodePromise
from ...epure import Epure


class Table(DbEntity):
    header:TableHeader
    if TYPE_CHECKING:
        parser: TermParser
    querying_proxy: QueryingProxy
    resource_proxy: DbProxy

    if TYPE_CHECKING:
        resource:TableStorage

    @property
    def db(self):
        return self.resource

    def __init__(self, name: str,
            header:Union[TableHeader, Dict[str, Any]]=None, resource:Resource=None, namespace:str = '') -> None:        
        check_type('header', header, [TableHeader, dict, NoneType])
        from ...parser.term_parser import TermParser

        self._set_header(header)
        super().__init__(name, namespace, resource)

        self.parser = TermParser(self)
        self.querying_proxy = TableProxy(self.db, self)
        self.resource_proxy = DbProxy(self.db)
        


    def _serialize(self, node: Savable) -> Dict[str, str]:
        res = {}
        for field_name, field_type in node.annotations.items():
            if isinstance(field_type, Constraint):
                field_type = field_type.py_type
                
            if node.is_excluded(field_name, field_type):
                continue
            if field_name not in self.header:
                continue
            if not hasattr(node, field_name):
                continue

            field_val = getattr(node, field_name, None)
            #working for db:
            if isinstance(field_val, Savable):
                field_type = field_val.annotations['node_id']
                field_val = field_val.save(True).node_id
            
            if (isinstance(field_type, type) and issubclass(field_type, Savable)\
                and not isinstance(field_val, Savable)):
                field_type = type(field_val)
                
            field_val = self.db.cast_py_db_val(field_val, field_type)

            res[field_name] = field_val
        
        return res



    def create(self, node: Node, asynch:bool=False) -> object:
        node.node_id = self.generate_id()
        script = self.serialize_for_create(node)
        if asynch:
            self.cache(script)
        else:
            self.execute(script)
        return node

    
    def read(self, *args, **kwargs) -> Any:        
        # if not (args or kwargs):
        #     return self.read_by_fields(**kwargs)
        
        if kwargs:
            return self.read_by_kwargs(*args, **kwargs)

        if args:
            selector = args[0]
        else:
            # args.__add__(self.querying_proxy)
            selector = self.querying_proxy@True
            args = args + (selector,)

        if isinstance(selector, str):
            return self.read_by_sql(selector)

        if callable(selector):
            return self.read_by_function(selector)

        if isinstance(args[-1], Term):
            selector = args[-1]
            return self.read_by_term(args[0:-1], selector)

        raise NotImplementedError(f'couldn read by object of type {type(selector)}')


    # def read_by_fields(self):
    #     raise NotImplementedError

    def read_by_kwargs(self, header:List[str]=[], operator:str="", **kwargs):
        term_header = []
        tp = self.querying_proxy
        for col_name in header:
            column = getattr(tp, col_name)
            term_header.append(column)
        
        if not term_header:
            term_header.append(tp)
        # from ...parser.leaf import Primitive 
        # term = term_header @ Primitive(True)
        kwargs_items = list(kwargs.items())
        first_item = kwargs_items[0]
        term = term_header @ getattr(tp, first_item[0]) == first_item[1]
        for (key, val) in kwargs_items[1:]:
            if operator == 'or':
                term = term | getattr(tp, key) == val
            else:
                term = term & getattr(tp, key) == val
        return self.read(term)

    def read_by_sql(self, selector):
        res = self.execute(selector)
        res = self.deserialize(res)
        return res

    def read_by_function(self, func):
        selector = func(self.querying_proxy, self.resource_proxy)
        if isinstance(selector, list) or isinstance(selector, tuple):
            return self.read(*selector)
        return self.read(selector)

    def read_by_term(self, header, selector:Term):
        sql = self.parser.parse(header, selector)
        res = self.read(sql)
        return res

        

    def deserialize(self, rows: dict, lazy_read:bool=True):
        res = []
        if not rows:
            return res
        full_name_epure_dict = self._column_full_name_epure_dict(rows[0])        
        for node_dict in rows:
            res_row = self._init_epures_row(node_dict, full_name_epure_dict, lazy_read)
            res.append(res_row)
        return res


    def _init_epures_row(self, node_dict, full_name_epure_dict, lazy_read) -> Dict[type, dict]:
        epure_attrs_dict = OrderedDict()
        for full_name, db_val in node_dict.items():
            column_tuple = full_name_epure_dict[full_name]
            epure_cls = column_tuple[0]
            column = column_tuple[1]
            if epure_cls not in epure_attrs_dict:
                epure_attrs_dict[epure_cls] = {}

            attrs = epure_attrs_dict[epure_cls]            
            # db_val = node_dict[full_name]
            if isinstance(column, TableColumn):
                field_type = column.py_type
                if isinstance(field_type, Constraint):
                    field_type = field_type.py_type
                val = self.db.cast_db_py_val(db_val, field_type)
                attrs[column.name] = val
            else:
                val = self.db.cast_db_py_val(db_val)
                attrs[column] = val

        res = []
        for epure_cls, attrs in epure_attrs_dict.items():
            epure_obj = self._init_epure(epure_cls, attrs, lazy_read)
            res.append(epure_obj)
        return res


    def _init_epure(self, epure_cls:Epure, attrs:dict, lazy_read):

        res = None
        if True in epure_cls.init_params:
            res = epure_cls(**attrs)
        else:
            arguments = {}
            for name in epure_cls.init_params:
                if isinstance(name, str) and name in attrs:
                    arguments[name] = attrs[name]
            try:
                res = epure_cls(**arguments)
            except Exception as ex:
                raise ex
        if res is None:
            raise DeserializeError
        
        for field_name, field_val in attrs.items():
            setattr(res, field_name, field_val)

        if not (hasattr(res, 'annotations') and 'node_id' in attrs):
            return res

        for field_name, field_type in res.annotations.items():

            if epure_cls.is_excluded(field_name):
                continue

            if field_name not in attrs:
                node_id = attrs['node_id']
                promise = FieldPromise(epure_cls.resource, node_id, field_name)
                setattr(res, field_name, promise)

            elif isinstance(field_type, Epure):
                node_id = attrs[field_name]
                if issubclass(field_type, Proto) and field_name == '__proto__':
                    proto = field_type.resource.read(node_id=node_id)
                    res.set_proto_fields(proto)
                elif lazy_read:
                    promise = NodePromise(field_type.resource, node_id)
                    setattr(res, field_name, promise)
                elif not lazy_read:
                    node = field_type.resource.read(node_id=node_id)
                    setattr(res, field_name, node)
        return res



    def _column_full_name_epure_dict(self, node_dict) -> Dict[str, tuple]:
        res = OrderedDict()
        for full_name, val in node_dict.items():
            dot_name = full_name.replace('___', '.')
            split = dot_name.rsplit('.', 1)
            table_name = split[0]
            field_name = split[1]            
            epure_cls = self.db.get_epure_by_table_name(table_name)
            if not epure_cls:
                res[full_name] = (object, field_name)
                continue

            epure_table = getattr(epure_cls, 'resource', None)
            if epure_table == None or not isinstance(epure_table, Table):
                raise DbError(f'epure {epure_cls} must have table as resourse')
            column = epure_table.header[field_name]
            res[full_name] = [epure_cls, column]
        return res


    def update(self, node: Savable, asynch:bool=False) -> object:
        script = self.serialize_for_update(node)
        if asynch:
            self.cache(script)
        else:
            self.execute(script)
        return node


    def cache(self, script: str):        
        self.db.cache_queue.append(script)


    def _set_header(self, header):
        if header == None:
            header = TableHeader(table=self)
        
        if isinstance(header, Dict):
            header = TableHeader(columns=header, table=self)
        
        self.header = header
        self.header.resource = self
       
    def serialize_for_create(self, node: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_read(self, node: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_update(self, node: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_delete(self, node: Savable, **kwargs) -> object:
        raise NotImplementedError

    
    def serialize_header(self, db: TableStorage=None, **kwargs) -> List[Dict[str, str]]:
        res: List[Dict[str, str]] = list()

        if db == None:
            db = self.db

        if db == None:
            raise EpureError('undefined db for header serialization')
        
        header = self.header
        for column_name in header:
            column = header[column_name]
           
            serialized = {
                "column_name": column.name,
                "column_type": column.serialize_type(db) #db.get_db_type(column.column_type)
            }
            res.append(serialized)
        return res

    def _add_node_id_fields(self, header:List[QueryingProxy]):
        res = []
        for item in header:
            if isinstance(item, ColumnProxy):
                tp = item.__table_proxy__
                node_id = getattr(tp, 'node_id', None)
                if node_id is not None and not node_id.in_header(header + res):
                    res.append(node_id)
        
        # res = tuple(res)
        header = tuple(header)
        
        if isinstance(header[0],str):
            res = set()
            for item in header:
                sp = item.split('.')

                if len(sp) == 3:
                    node_id_column = f'{sp[0]}.{sp[1]}.node_id'
                    if node_id_column not in header:
                        res.add(node_id_column)
                    
        res = tuple(res)



        return header + res


    # def get_column_header_name(self, column_name:str):
    #     raise NotImplementedError

        # tables = []
        # columns = []
        # for item in header:
        #     if isinstance(item, TableProxy):
        #         table_name = item.str(False, True)
        #         tables.append(table_name)
        #     if isinstance(item, ColumnProxy):
        #         column_name = item.str(False, True)
        #         columns.append(column_name)

        # res = []
        # for item in header:
        #     if isinstance(item, ColumnProxy):
        #         tp = item.__table_proxy__
        #         table_name = tp.str(False, True)
        #         if (not table_name in tables) and\
        #                 hasattr(tp, 'node_id') and\
        #                 tp.node_id.str(False, True) not in columns:
        #             res.append(tp.node_id)
        #             columns.append(tp.node_id.str(False, True))
        # return res


from .binary import Binary
from ast import BinOp, Compare
from ..errors import UserInputError


class BinOperation(Binary, BinOp):
    def __init__(self, left, right, operator='') -> None:        
        from .leaf import TableProxy
        super().__init__(left, right, operator)

        left = self.left
        right = self.right

        if self.is_join() and isinstance(right, TableProxy):
            raise UserInputError('right operand of join must be logical expression, not table')
        

        if isinstance(left, Comparison) and isinstance(right, Comparison):
            self.set_parentheses()
        elif isinstance(left, Comparison) and isinstance(right, BinOperation)\
            and right.is_correct() and not right.is_join():
            self.set_parentheses()
        elif isinstance(left, BinOperation) and isinstance(right, Comparison)\
            and left.is_correct() and not left.is_join():
            self.set_parentheses()
        
    def is_join(self):
        from .leaf import TableProxy
        if self.operator not in ('<<', '>>'):
            return False
        res = isinstance(self.left, TableProxy)
        return res


    def is_correct(self):
        res = self.left_closed() and self.right_closed()
        return res

    def left_open(self):
        left_open = not self.left_closed()
        return left_open

    def right_open(self):
        right_open = not self.right_closed()
        return right_open

    def left_closed(self):
        from .leaf import TableProxy
        left_leaf = self.get_last_left_term()
        left_closed = isinstance(left_leaf, TableProxy) or isinstance(left_leaf.right_parent, Comparison)
        return left_closed

    def right_closed(self):
        from .leaf import TableProxy
        right_leaf = self.get_last_right_term()
        right_closed = isinstance(right_leaf, TableProxy) or isinstance(right_leaf.left_parent, Comparison)
        return right_closed



class Comparison(Binary, Compare):
    def __init__(self, left, right, operator='') -> None:
        from .leaf import ColumnProxy
        from .leaf import Primitive
        from .leaf import Leaf        

        super().__init__(left, right, operator)

        left = self.left
        right = self.right

        if self.is_join():
            self.set_join_parentheses()        
        

        if isinstance(left, ColumnProxy) and isinstance(right, ColumnProxy):
            self.set_parentheses()
        elif isinstance(left, ColumnProxy) and isinstance(right, Primitive):
            self.set_parentheses()
        elif isinstance(left, Primitive) and isinstance(right, ColumnProxy):
            self.set_parentheses() #raize Error?
        elif isinstance(left, Primitive) and isinstance(right, Primitive):
            self.set_parentheses()

        elif isinstance(left, BinOperation) and isinstance(right, Leaf):            
            # if left.is_correct():
            #     self.set_parentheses()
            if left.right_open() and left.left_closed():
                self.set_parentheses()

        elif isinstance(left, Leaf) and isinstance(right, BinOperation):            
            # if right.is_correct():
            #     self.set_parentheses()
            if right.left_open() and right.right_closed():
                self.set_parentheses()

        elif isinstance(left, BinOperation) and isinstance(right, BinOperation):
            if left.is_correct() and right.is_correct():
                self.set_parentheses()
            elif left.left_closed() and left.right_open() \
                and right.left_open() and right.right_closed():
                self.set_parentheses()

        
    def is_join(self):
        if not isinstance(self.left, Binary):
            return False
        left_right_leaf = self.left.get_last_right_term()
        x = left_right_leaf.left_parent
        res = hasattr(x, 'operator') and x.operator in ('<<', '>>')
        return res

    def set_join_parentheses(self):
        from .leaf import TableProxy
        right = self.right
        right_left_leaf = right
        left_right_leaf = self.left.get_last_right_term()
        if isinstance(right, Binary):
            right_left_leaf = right.get_last_left_term()
        if isinstance(right_left_leaf, TableProxy):
            raise UserInputError('right operand of join must be logical expression, not table')
        left_right_leaf.left_parentheses_count += 1        
        right_left_leaf.right_parentheses_count += 1


from configparser import ConfigParser, RawConfigParser
from .file_storage import FileStorage
from typing import Any
from io import StringIO
from .ini_section import ActualIniSection, NoneIniSection


class IniFile(FileStorage):

    parser: RawConfigParser
    virtual_section: ActualIniSection

    def __init__(self, path:str):

        parser = self.parser = ConfigParser()
        with open(path, 'r') as file:
            config_string = '[VIRTUAL_SECTION]\n' + file.read()            
            parser.read_string(config_string)
        self.virtual_section = ActualIniSection(parser, 'VIRTUAL_SECTION')

        
    def __getattr__(self, attr_name: str) -> Any:
        if attr_name in self.virtual_section:
            return self.virtual_section.__getattr__(attr_name)
        elif attr_name in self.parser:
            return ActualIniSection(self.parser, attr_name)
        else:
            return NoneIniSection(self.parser, attr_name)

    # def save(self) -> Any:
    #     output = StringIO()
    #     self.parser.write(output)
    #     res = output.getvalue().split('\n', 1)[2]
    #     with open(self.path, 'w') as configfile:
    #         configfile.write(res)


from types import NoneType
from typing import TYPE_CHECKING, Dict, Union, List, ItemsView, Any, Type, Callable, cast
from ...helpers.type_helper import check_subclass
from ...errors import EpureError

__all__ = ['Constraint', 'Default', 'NotNull', 'Uniq', 'Prim', 'Foreign', 'Check']

class Constraint(type):
    py_type:type = NoneType
    __origin__:type = NoneType
    
    def __getitem__(self:Type, params:Any):
        res = self.__class__(self.__name__, self.__bases__, dict(self.__dict__))
        res.__origin__ = self
        if isinstance(params, type):
            res.py_type = params
        if isinstance(params, tuple):
            res.py_type = params[0]
            res.set_params(*params[1:])

        if issubclass(res.py_type, Constraint):
            # ( hasattr(res.py_type, '__origin__') and  issubclass(res.py_type.__origin__, Constraint)):
            raise EpureError('Constraint parameter cant be Constraint')
        return res

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if self.py_type != other.py_type:
            return False
        if self.__origin__ != other.__origin__:
            return False
        if self.__name__ != other.__name__:
            return False
        
        if self.__origin__ == Default:
            self = cast(Default, self)
            other = cast(Default, other)
            return self.default == other.default
        return True

class Default(metaclass=Constraint):
    py_type:type = NoneType
    default:Any = None

    @classmethod
    def set_params(cls, default:Any):
        cls.default = default



class NotNull(Default):
    pass
    # py_type:type
    # default:Any
    # def __class_getitem__(cls, py_type:type, default:Any=None):
    #     cls.py_type = py_type
    #     cls.default = default
    #     return cls

class Uniq(metaclass=Constraint):
    pass

class Prim(NotNull, Uniq):
    pass

class Foreign(metaclass=Constraint):
    table:str = ''
    column:str = ''
    
    @classmethod
    def set_params(cls, table:str, column:str):        
        cls.table = table
        cls.column = column


class Check(metaclass=Constraint):    
    condition:Any = None
    
    @classmethod
    def set_params(cls, condition:Any):
        cls.condition = condition
        return cls



# https://www.postgresql.org/docs/current/sql-createtable.html

# where column_constraint is:

# [ CONSTRAINT constraint_name ]
# { NOT NULL |
#   NULL |
#   CHECK ( expression ) [ NO INHERIT ] |
#   DEFAULT default_expr |
#   GENERATED ALWAYS AS ( generation_expr ) STORED |
#   GENERATED { ALWAYS | BY DEFAULT } AS IDENTITY [ ( sequence_options ) ] |
#   UNIQUE index_parameters |
#   PRIMARY KEY index_parameters |
#   REFERENCES reftable [ ( refcolumn ) ] [ MATCH FULL | MATCH PARTIAL | MATCH SIMPLE ]
#     [ ON DELETE referential_action ] [ ON UPDATE referential_action ] }
# [ DEFERRABLE | NOT DEFERRABLE ] [ INITIALLY DEFERRED | INITIALLY IMMEDIATE ]


#Default = DEFAULT
#NotNull = NOT NULL, need default
#Uniq = UNIQUE, can be nullable
#Foreign = REFERENCES, must refer to Prim or Uniq

#Prim = PRIMARY KEY = Uniq + NotNull need method @Prim, generated unique Prim by self
#Check = CHECK


from ..savable import Savable
from ...named import NamedByClass
from abc import abstractmethod
from ..resource import Resource

class DbEntity(Savable, NamedByClass):
    name = ''
    namespace = ''

    @abstractmethod
    def __init__(self, name:str='', namespace:str='', resource:Resource=None) -> None:
        self.name = name if name else self.__class__.__name__
        self.namespace = namespace
        super().__init__(resource)

    # @property
    # def full_name(self):
    #     if hasattr(self, 'namespace') and self.namespace:
    #         return self.namespace + '.' + self.name
    #     if not (hasattr(self, 'name') and self.name):
    #         if isinstance(self, type):
    #             self.name = self.__name__
    #         else:
    #             self.name = self.__class__.__name__
    #     return self.name


from __future__ import annotations
from types import NoneType
from typing import cast, Callable, List, Any, get_type_hints, Dict, TYPE_CHECKING
from .resource.resource import Resource
from .helpers.type_helper import check_type
from .errors import EpureError, DefaultConstraintError
from .resource.db.constraint import Foreign, Default, Constraint


# from types import FunctionType
from .resource.node.node import TableNode
from .resource.savable import Savable
if TYPE_CHECKING:
    from .resource.db.table_storage import TableStorage
    from .resource.db.table import Table
from .resource.db.db_entity import DbEntity
import functools
from .parser.term import Term
from .resource.node.proto import Proto
from inspect import signature



class Epure(type, Savable):
    
    EDb:TableStorage
    epures:List[Epure] = []
    # annotations:Dict[str,Any]
    resource:Savable
    prepared_resource:object
    init_params:object
    
    # def __new__(mcls, name: str, cls_bases: Tuple[type, ...], namespace: dict[str, Any]) -> Epure:
    #     return super().__new__(mcls, name, cls_bases, namespace)    
    
    def __call__(self, *args, **kwargs):
        if not self.is_saved:
            self.save_epure()
            self.is_saved = True    # __getattr__ is called here, you cannot enter __getattr__ with debuger, so put breakpoint on start of getattr func

        res = super(Epure, self).__call__(*args, **kwargs)
        return res


    def __getattr__(self, attr_name: str) -> Any:
        if not self.is_saved:
            self.is_saved = True    # __getattr__ is called here, you cannot enter __getattr__ with debuger, so put breakpoint on start of getattr func
            self.save_epure()
            return getattr(self, attr_name)
        raise AttributeError(f"'{type(self)}' object has no attribute '{attr_name}'")
        # if self.is_saved:
        #     raise AttributeError(f"'{type(self)}' object has no attribute '{attr_name}'")
        # self.is_saved = True
        # self.save_epure()        
        # return getattr(self, attr_name)

    #decorator
    @classmethod
    def read(cls, method):
        querying_proxy = None
        resource_proxy = None
        if getattr(cls, 'resource', None):
            querying_proxy = getattr(cls.resource, 'querying_proxy', None)
            resource_proxy = getattr(cls.resource, 'resource_proxy', None)

        @functools.wraps(method)
        def wrap(self, *args, **kwargs):
            res = None
            if querying_proxy and resource_proxy:
                res = method(self, querying_proxy, resource_proxy, *args, **kwargs)
            if querying_proxy:
                res = method(self, querying_proxy, *args, **kwargs)
            if resource_proxy:
                res = method(self, resource_proxy, *args, **kwargs)
            else:
                res = method(self, *args, **kwargs)
            if isinstance(res, Term):
                return self.resource.read(res)
            return res
        return wrap
        # def reader(self:Table, *args, **kwargs):
        #     pseudo_self = PresudoTable(self)
        #     pseudo_db = PseudoDb(self.db)
        #     script = selector(pseudo_self, pseudo_db, *args, **kwargs)
        #     return self.execute(selector)
        # setattr(self, selector.__name__, reader)

    def prepare_save(self, resource:object):
        self.__class__.epures.append(self)
        self.prepared_resource = resource


    def save_epure(self):
        self.set_resource()
        self.init_params = self.get_init_params()

    def get_init_params(self):
        res = []
        sig = signature(self.__init__)
        for key, val in sig.parameters.items():
            if val.kind == val.VAR_POSITIONAL:
                    continue
            if val.kind == val.VAR_KEYWORD:
                res.append(True)
            else:
                res.append(key)
        return res[1:]
        # sig = signature(epure_cls.__init__)
        # params = sig.parameters
        # params_vals = params.values()
        # has_kwargs = any([True for p in params_vals if p.kind == p.VAR_KEYWORD])
        # if has_kwargs:
        #     res = epure_cls(**kwargs)
        # else:
        #     arguments = {}
        #     for key, val in params:
        #         if val.kind == val.VAR_POSITIONAL:
        #             continue
        #         if key in kwargs:
        #             arguments[key] = kwargs[key]
        #             continue
        #         if val.default == val.empty:
        #             raise DeserializeError
        #     res = epure_cls(**arguments)


    def set_resource(self):
        resource = self.prepared_resource
        check_type('resource', resource, [str, Savable, NoneType])

        if type(resource) == Savable:
            resource = cast(Savable, resource)
            if not hasattr(resource, 'resource'):
                raise EpureError('For using savable as resource for epure, savable must have own resource')
            grandpa = resource.resource
            grandpa.update(self)
        else:
            table_name = self._get_table_name(self, resource)
            resource = self._create_or_update(table_name)

        self.resource = resource

        

    def _create_or_update(self, table_name:str) -> Table:
        from .resource.db.table import Table
        table:Table = self._get_table_by_cls(table_name)
        res:DbEntity
        if table_name in self.EDb:
            deleted_columns = self.EDb[table_name].header.deleted_columns
            table.header.deleted_columns = deleted_columns
            res = self.EDb.update(table)
        else:
            res = self.EDb.create(table)
        res = cast(Table, res)
        return res

    def _get_table_name(self, cls, resource: object) -> str:
        
        if not isinstance(resource, str):
            resource = cast(Savable, resource)
            return resource.full_name

        if resource:
            return resource
        
        return self.EDb._get_full_table_name(cls.__name__).full_name

        

    def _get_table_by_cls(self, table_name: str = '') -> Table:
        from .resource.db.table import Table
        table:Table
        
        full_name = self.EDb._get_full_table_name(table_name)        
        columns = {}

        for field_name, py_type in self.annotations.items():
            py_type = cast(type, py_type)
            if self.is_excluded(field_name, py_type):
                continue
            py_type = self.get_py_type(field_name, py_type)
            
            columns[field_name] = py_type

        if len(columns) == 1 and list(columns.keys())[0] == 'node_id':
            TableCls = self.EDb.nosql_table_type
        else:
            TableCls = self.EDb.default_table_type
        table = TableCls(full_name.name, columns, self.EDb, full_name.namespace)
        return table

    def get_py_type(self, field_name:str, py_type:type) -> type:
       
        if py_type in self.epures:
            py_type = cast(Epure, py_type)
            return self.get_py_type(field_name, py_type.annotations['node_id'])
            
        if isinstance(py_type, Constraint) and issubclass(py_type.__origin__, Default):
            return self.get_default_type(field_name, py_type)

        return py_type

    def get_default_type(self, field_name:str, py_type:Default) -> Any:
        default = getattr(self, field_name, None)
        if not (default or py_type.default):
            raise DefaultConstraintError(self.full_name, field_name)
        if not py_type.default:
            py_type.default = default

        if not isinstance(py_type.default, py_type.py_type):
            raise DefaultConstraintError(message=f'''field {field_name} of class {self.full_name} 
                    has wrong defaul value {py_type.default} not corresponding to {py_type.py_type} type''')

        return py_type


    def get_foreign_type(self, foreign:Epure) -> Any:
        foreign_id_type = foreign.annotations['node_id']

        foreign_table = self._get_table_name(foreign, foreign.prepared_resource)

        foreign_column = 'node_id'
        column_type = Foreign[foreign_id_type, foreign_table, foreign_column]
        return column_type


        

def epure(resource:object='', saver:type=TableNode, epure_metaclass:type=Epure) -> Callable:
    check_type('resource', resource, [str, Savable, NoneType])

    def epure_creator(cls:type) -> Epure:
        epure_cls = _create_epure(cls, saver, epure_metaclass)
        epure_cls.is_saved = False      # might not change true to false in the debuger, because this is overrided setter.
        epure_cls.prepare_save(resource)

        del cls
        return epure_cls

    return epure_creator


def _create_epure(cls, saver, _Epure):
    cls_dict = dict(cls.__dict__)
    if issubclass(cls, Savable):
        return _Epure(cls.__name__, cls.__bases__, cls_dict)

    bases = list(cls.__bases__)
    if object in bases:
        bases.remove(object)
    bases.append(saver)
    return _Epure(cls.__name__, tuple(bases), cls_dict)


def proto(resource:object='', saver:type=Proto, epure_metaclass:type=Epure) -> Callable:
    return epure(resource, saver, epure_metaclass)


from uuid import UUID
from ..savable import Savable
from ..resource import Resource
from ...errors import ResourceException
from typing import Any, Dict
import jsonpickle
from ..db.constraint import Constraint

class Node(Savable):

    node_id:object
    __exclude__:list

    def __init__(self, node_id:object=None, resource:Resource=None) -> None:
        if node_id != None:
            self.node_id = node_id
        super().__init__(resource)

    @classmethod
    def from_dict(_cls, _dict:Dict[str, Any])->object:
        raise NotImplementedError

    def save(self, asynch:bool=False):
        if not hasattr(self, 'resource'):
            raise ResourceException('Unable save Savable, resource not found')
        resource = self.resource
        
        
        if hasattr(self, 'node_id') and self.node_id:
            res = resource.update(self, asynch)
            return res
        else:
            res = resource.create(self, asynch)
            return res


class TableNode(Node):
    db:Resource
    node_id: UUID
    resource:Savable

    def __init__(self, node_id:object=None, resource:Resource=None) -> None:
        if node_id != None and isinstance(node_id, str):
            node_id = UUID(node_id)
        super().__init__(node_id, resource)

    @property
    def table(self) -> Savable:
        return self.resource

    @classmethod
    def from_dict_deep(_cls, _dict:Dict[str, Any])->object:

        # instance = _cls.__call__()
        instance = _cls()
        
        for field_name, val in _dict.items():
            if field_name in instance.annotations:
                _type = instance.annotations[field_name]
                if isinstance(_type, Savable):
                    val = _type(val)
                setattr(instance, field_name, val)


        return instance
    
    @classmethod
    def from_dict(_cls, _dict:Dict[str, Any])->object:

        # instance = _cls.__call__()
        instance = _cls()
        
        for field_name, val in _dict.items():
            if field_name in instance.annotations:
                setattr(instance, field_name, val)


        return instance
    
    def to_json(self):
        from ..db.table import NodePromise
        res = {}
        for field_name, field_type in self.annotations.items():
            if isinstance(field_type, Constraint):
                field_type = field_type.py_type
                
            if self.is_excluded(field_name, field_type):
                continue
            if field_name not in self.table.header:
                continue
            if not hasattr(self, field_name):
                continue

            field_val = getattr(self, field_name, None)

            if isinstance(field_val, NodePromise):
                field_val = getattr(field_val, 'node_id2', None)
                # field_val = getattr(field_val, 'none2', None)

            #working for db:
            if isinstance(field_type, Savable) and not isinstance(field_val, UUID):
                field_val = getattr(self, field_name, None)
                field_type = field_val.annotations['node_id']
                field_val = field_val.save(True).node_id

            if isinstance(field_val, UUID):
                field_val = str(field_val)

            res[field_name] = field_val

        res = jsonpickle.encode(res)

        return res

    __exclude__:list = Node.__exclude__ + ['table', 'db']


from .file_storage import FileStorage
from ..savable import Savable
import jsonpickle
import json

class JsonFile(FileStorage):
    
    def serialize_for_update(self, savable: Savable, **kwargs) -> object:
        res = str(jsonpickle.encode(savable))
        return res

    def deserialize(self, json_obj: str, **kwargs) -> Savable:
        res = json.dumps(json_obj)
        res = jsonpickle.decode(res)
        return res


from __future__ import annotations
from types import LambdaType, NoneType
from typing import TYPE_CHECKING, Dict, Union, List, ItemsView, Any, Type, Callable, cast, Optional


from .constraint import NotNull

if TYPE_CHECKING:
    from .table import Table
    from .table_storage import TableStorage
from ..savable import Savable
from ...helpers.type_helper import check_type
from ...errors import EpureError
# from uuid import uuid4
from .table_column import TableColumn
from ..db.db_entity_resource import DbEntityResource
from .constraint import Constraint, NotNull, Default, Prim, Foreign, Uniq, Check
import uuid


class TableHeader(Savable):
    columns:Dict[str,TableColumn]
    deleted_columns:Dict[str, TableColumn]

    if TYPE_CHECKING:
        table:Table

    def __init__(self, 
            columns:Dict[str, type]=None, table:Table=None,
            name: str = '') -> None:
        check_type('columns', columns, [dict, NoneType])

        if table:
            self.table = table
        self.columns = {}
        self.deleted_columns = {}
        super().__init__(name)
        if not columns:
            return

        for name, py_type in columns.items():            
            column = TableColumn(name, py_type)
            column = cast(TableColumn, column)
            self.columns[name] = column

    @property
    def db(self):
        return self.table.db

    def __getitem__(self, key:str):
        return self.columns[key]

    def __iter__(self):
        res = self.columns.__iter__()
        return res

    def __len__(self):
        return self.columns.__len__()

    def __sub__(self, other: TableHeader) -> List[TableColumn]:
        res = []
        for column_name in self:
            self_column = self[column_name]

            if self_column.is_deleted:
                continue

            if self_column not in other:
                    res.append(self_column)
        return res

    def __contains__(self, other:Any) -> bool:
        if not isinstance(other, TableColumn):
            return other in self.columns
        column = cast(TableColumn, other)

        if column.is_deleted:
            return False

        if column.name not in self:
            return False

        self_column = self[column.name]

        if column == self_column:
            return True

        if isinstance(self_column.py_type, Constraint) or isinstance(column.py_type, Constraint):
            return False

        return self.db.same_db_type(self_column.py_type, column.py_type)

    def create(self, column: Savable) -> Any:
        check_type('column', column, TableColumn)

        script = ''
        restored_col_name = ''
        column_is_for_restore = False

        for restored_col_name, restored_column in self.deleted_columns:
            if restored_col_name[:-4] == f'{column.name}___del' and restored_column.py_type == column.py_type:
                script = self.serialize_for_restore(column)
                column_is_for_restore = True
                break

        if script == '':
            script = self.serialize_for_create(column)

        script = str(script)
        self.execute(script)

        if column_is_for_restore:
            self.deleted_columns.pop(restored_col_name)

        self.columns[column.name] = column

        return column

    def delete(self, column: Savable) -> Any:
        check_type('column', column, TableColumn)

        random_id = str(uuid.uuid4())[0:4]
        del_column_name = f'{column.name}___del{random_id}'

        script = self.serialize_for_delete(column, del_column_name)
        script = str(script)
        self.execute(script)

        self.columns.pop(column.name)
        self.deleted_columns[del_column_name] = column
        column.name = del_column_name

        return column

    def update(self, py_column: Savable) -> Any:
        check_type('py_column', py_column, TableColumn)

        py_column = cast(TableColumn, py_column)
        db_column = self.columns[py_column.name]

        if py_column in self:
            return db_column

        random_id = str(uuid.uuid4())[0:4]
        del_column_name = f'{py_column.name}___del{random_id}'

        script = ''
        restored_col_name = ''
        column_is_for_restore = False

        for restored_col_name, restored_column in self.deleted_columns.items():
            if restored_col_name[:-4] == f'{py_column.name}___del' and restored_column.py_type == py_column.py_type:

                script = self.serialize_for_restore(py_column, restored_col_name, del_column_name)
                column_is_for_restore = True
                break
        
        if script == '':
            script = self.serialize_for_update(py_column, del_column_name)
        script = str(script)
        self.execute(script)

        db_column.name = del_column_name
        self.deleted_columns[del_column_name] = db_column

        if column_is_for_restore:
            self.deleted_columns.pop(restored_col_name)
            
        self.columns[py_column.name] = py_column

        return db_column

    def serialize_for_create(self, column: Savable, **kwargs) -> object:
        check_type('column', column, TableColumn)
        column = cast(TableColumn, column)
        table_name = self.table.full_name

        db:DbEntityResource
        if 'db' in kwargs:
            db = kwargs['db']
        else:
            db = self.table.db        

        create_script = self._create_column_script(table_name, column, db)
        return create_script

    def serialize_for_update(self, column: Savable, del_column_name:str, **kwargs) -> object:
        check_type('column', column, TableColumn)
        column = cast(TableColumn, column)
        table_name = self.table.full_name

        pre_delete_script = self._serialize_pre_delete(table_name, column, del_column_name)

        db:DbEntityResource
        if 'db' in kwargs:
            db = kwargs['db']
        else:
            db = self.table.db

        create_script = self._create_column_script(table_name, column, db)

        migrate_script = ''
        if self.table.db.migrate_on_delete:
            migrate_script = self._migrate_columns_script(table_name, del_column_name, column.name)

        return pre_delete_script + create_script + migrate_script

    def serialize_for_delete(self, column: Savable, del_column_name:str, **kwargs) -> object:
        check_type('column', column, TableColumn)
        column = cast(TableColumn, column)
        table_name = self.table.full_name

        pre_delete_script = self._serialize_pre_delete(table_name, column, del_column_name)
        return pre_delete_script
    
    def serialize_for_restore(self, column:Savable, restore_col_name:str, del_column_name:str=None):
        check_type('column', column, TableColumn)
        restore_script = ''
        script = ''
        pre_delete_script = ''

        table_name = self.table.full_name
        column = cast(TableColumn, column)
        restore_script = self._serialize_for_restore(table_name, column, restore_col_name)
        if del_column_name:
            pre_delete_script = self._serialize_pre_delete(table_name, column, del_column_name)
        script = pre_delete_script + restore_script

        return script

    def _serialize_for_restore(self, table_name:str, column:TableColumn, restore_name) -> object:
        script = self._rename_column_script(table_name, restore_name, column.name)
        return script

    def _serialize_pre_delete(self, table_name:str, column:TableColumn, del_column_name:str) -> str:
        script = self._rename_column_script(table_name, column.name, del_column_name)\
                + self._set_nullable_script(table_name, del_column_name)
        return script

    def deserialize(self, column_dict: object, **kwargs) -> Savable:
        check_type('column_dict', column_dict, [dict, list])
        column_dict = cast(dict, column_dict)

        column_dict = self._deserialize(column_dict)

        column_name = column_dict['column_name']
        db_type = column_dict['db_type']

        not_null = column_dict['not_null']
        default = column_dict['default']
        uniq = column_dict['uniq']
        prim = column_dict['prim']
        foreign = column_dict['foreign']
        

        db = self._get_db(**kwargs)
        try:
            column_type = db.get_py_type(db_type)
            if prim:
                column_type = Prim[column_type]
            elif uniq:
                column_type = Uniq[column_type]
            elif not_null:
                column_type = NotNull[column_type, default]
            elif default:
                column_type = Default[column_type, default]
            elif foreign:                
                foreign_table = column_dict['foreign_table']
                foreign_column = column_dict['foreign_column']
                column_type = Foreign[column_type, foreign_table, foreign_column]
            

        except KeyError as ex:
            raise TypeError(f'{db_type} is unknown for {db.name} '
                            'set db_py_types attribute properly') from ex       
        
        return TableColumn(column_name, column_type)

    def _deserialize(self, column_dict:dict) -> dict:
        return column_dict

    def _get_db(self, **kwargs) -> TableStorage:

        db = None
        if 'db' in kwargs:
            db = kwargs['db']
            return db
        
        if hasattr(self, 'table') and hasattr(self.table, 'db'):
            db = self.table.db

        if db is None:
            raise EpureError('undefined db for column serialization')
        return db

    def _rename_column_script(self, table_name:str, old_name:str, new_name:str) -> str:
        raise NotImplementedError
        
    def _set_not_null_script(self, table_name:str, column_name:str) -> str:
        raise NotImplementedError

    def _set_nullable_script(self, table_name:str, column_name:str) -> str:
        raise NotImplementedError

    def _create_column_script(self, table_name:str, column:TableColumn, db:DbEntityResource) -> str:
        raise NotImplementedError

    def _migrate_columns_script(self, table_name:str, from_column:str, to_column:str) -> str:
        raise NotImplementedError

    def _set_column(self, column:TableColumn):
        check_type('column', column, TableColumn)
        if column.is_deleted:
            self.deleted_columns[column.name] = column
        else:
            self.columns[column.name] = column
    
    def serialize_read_column(self, column:TableColumn, full_names:bool):
        raise NotImplementedError


from .resource import Resource

class NodePromise:
    resource:Resource
    node_id:object
    def __init__(self, resource, node_id) -> None:
        self.resource = resource
        self.node_id = node_id

    def __getattr__(self, obj, objtype=None):
        return self.resource.read(node_id=self.node_id)


class FieldPromise(NodePromise):
    field_name:str
    def __init__(self, resource, node_id, field_name) -> None:
        self.field_name = field_name
        return super().__init__(resource, node_id)

    def __getattr__(self, obj, objtype=None):
        res = self.resource.read([self.field_name], node_id=self.node_id)
        return getattr(res, self.field_name)


from __future__ import annotations
from typing import Dict, List, TYPE_CHECKING
from uuid import UUID, uuid4
from ..errors import EpureParseError
if TYPE_CHECKING:
    from ..resource.db.db import Db


class Term:
    id:UUID
    terms_graph: Dict[UUID, Term]
    left_parent:Term = None
    right_parent:Term = None
    val:str
    __header__:list

    def __init__(self) -> None:
        self.id = str(uuid4())
        self.terms_graph = {self.id: self}
        self.__header__ = []

    #primitive operators:
    def __and__(self, other): #&
        return self.operation(self, other, 'and')

    def __rand__(self, other): #&
        return self.operation(other, self, 'and')

    def __or__(self, other:Term): #|
        return self.operation(self, other, 'or')

    def __ror__(self, other:Term): #|
        return self.operation(other, self, 'or')
    
    def __mod__(self, other): #%
        return self.operation(self, other, '%')

    #non primitive operators:
    def __xor__(self, other:Term): #^
        return self.operation(self, other, '^')

    def __rxor__(self, other:Term): #^
        return self.operation(other, self, '^')
    
    def __matmul__(self, other): #@
        return self.operation(self, other, '@')
    
    def __rmatmul__(self, other): #@
        return self.operation(other, self, '@')

    def __lshift__(self, other:Term): #<<
        return self.operation(self, other, '<<')
        # return self.join_operation(other, 'left')

    def __rshift__(self, other:Term): #<<
        return self.operation(self, other, '>>')

    #comparators:
    def __eq__(self, other): #==
        return self.comparison(self, other, '==')

    def __lt__(self, other): #<
        return self.comparison(self, other, '<')

    def __le__(self, other): #<=
        return self.comparison(self, other, '<=')

    def __ne__(self, other): #!=
        return self.comparison(self, other, '!=')

    def __gt__(self, other): #>
        return self.comparison(self, other, '>')

    def __ge__(self, other): #>=
        return self.comparison(self, other, '>=')
    
    # def __contains__(self, other): #in
    #     return self.comparison(self, other, 'in')



    def operation(self, left:Term, right:Term, operator:str):
        from .bin_operation import BinOperation
        if isinstance(right, TermHeader):
            right = right.val
        # if isinstance(right, list):
        #     if not isinstance(left, Term):
        #         from .leaf import Primitive
        #         left = Primitive(left)
        #     left.__header__ = self.merge_headers(left.__header__, right)
        #     return left
            
        # if isinstance(left, TermHeader):
        #     left = left.val
        # if isinstance(left, list):
        #     if not isinstance(right, Term):
        #         from .leaf import Primitive
        #         right = Primitive(right)
        #     right.__header__ = self.merge_headers(left, right.__header__)
        #     return right
        res = BinOperation(left, right, operator)
        res.merge_graphs()
        return res


    def comparison(self, left:Term, right:Term, operator:str):
        from .bin_operation import Comparison
        res = Comparison(left, right, operator)
        res.merge_graphs()
        return res

    def str(self, parentheses=False, full_names=False, translator:function=None):
        return self.serialize(parentheses, full_names, translator)

    def serialize(self, parentheses=True, full_names=True, for_header=False, translator=None) -> str:
        raise NotImplementedError
        

    def _simple_copy_terms(self, terms: List[Term]=None, translator=None):
        if terms is None:
            terms = list(self.terms_graph.values())

        copies = []
        for term in terms:
            term_copy = term._simple_copy(translator)
            copies.append(term_copy)

        for copy in copies:
            copy._restore_links(terms, copies)

        return copies



    def _simple_copy(self, translator=None) -> Term:
        res = Term()
        if hasattr(self, 'id') and self.id:
            res.id = self.id

        res.val = self.serialize(False, False, translator=translator)

        if hasattr(self, 'parentheses'):
            res.parentheses = self.parentheses
        else:
            res.parentheses = False
        return res


    def _restore_links(self, origin_terms, copies):
        self_orig_index = self.index_of(origin_terms)
        orig = origin_terms[self_orig_index]

        if hasattr(orig, 'left') and orig.left:
            left_copy_index = orig.left.index_of(copies)
            if left_copy_index != None:
                left_copy = copies[left_copy_index]
                self.left = left_copy
        if hasattr(orig, 'right') and orig.right:
            right_copy_index = orig.right.index_of(copies)
            if right_copy_index != None:
                right_copy = copies[right_copy_index]
                self.right = right_copy
        if hasattr(orig, 'left_parent') and orig.left_parent:
            left_parent_copy_index = orig.left_parent.index_of(copies)
            if left_parent_copy_index != None:
                left_parent_copy = copies[left_parent_copy_index]
                self.left_parent = left_parent_copy
        if hasattr(orig, 'right_parent') and orig.right_parent:
            right_parent_copy_index = orig.right_parent.index_of(copies)
            if right_parent_copy_index != None:
                right_parent_copy = copies[right_parent_copy_index]
                self.right_parent = right_parent_copy



    def index_of(self, terms: List[Term]):
        indexes = [index for (index, item) in enumerate(terms) if item.id == self.id]
        if len(indexes) > 1:
            raise EpureParseError('term occurred in a list more then once')
        if len(indexes) == 0:
            return None
        return indexes[0]


    def in_list(self, terms: List[Term]=None) -> bool:
        return self.index_of(terms) != None


    def go_until_hasattr(self, first_attr: str, second_attr: str, terms: List[Term]=None) -> Term:
        if terms is None:
            terms = list(self.terms_graph.values())

        iterat = self
        while True:
            first_attr_val: Term = getattr(iterat, first_attr, None)
            second_attr_val: Term = getattr(iterat, second_attr, None)

            if first_attr_val and first_attr_val.in_list(terms):
                iterat = first_attr_val

            elif second_attr_val and second_attr_val.in_list(terms):
                iterat = second_attr_val

            else:
                break

        return iterat


    def merge_headers(self, left_header, right_header):
        left_header = list(left_header)
        right_header = list(right_header)
        res = []
        for qp in right_header:
            if not qp.in_header(left_header):
                res.append(qp)
        res = left_header + res
        return res


class TermHeader(Term):
    def __init__(self, val:list) -> None:
        self.val = val


    # def go_until_hasattr(self, attr_names: List[str], terms: List[Term]=None) -> Term:
    #     next = self
    #     while True:
    #         has_attr = False
    #         for attr in attr_names:                
    #             if hasattr(next, attr):
    #                 tmp = getattr(next, attr) 
    #                 if tmp and tmp.index_of(terms) != None:
    #                     next = tmp
    #                     has_attr = True
    #                     break
    #         if not has_attr:
    #             return next

    #     # for attr_name in attr_names:
    #     #     if hasattr(self, attr_name):
    #     #         next = getattr(self, attr_name)
    #     #         break

    #     # if not next:
    #     #     return self
    #     # return next.go_until_hasattr(terms, attr_names)


from ..errors import EpureParseError
from .term import Term
from ast import BinOp, Compare
from .leaf import Primitive, QueryingProxy
from typing import Dict, List, TYPE_CHECKING

class Binary(Term):
    left:Term
    right:Term
    operator:str
    parentheses = False
    debugger = None
    val = None


    def __init__(self, left, right, operator='') -> None:

        if not (isinstance(left, Term) or isinstance(right, Term)):
            raise EpureParseError('expression is incorrect')

        if not isinstance(left, Term):
            left = Primitive(left)
        elif isinstance(left, QueryingProxy) and not left.__is_copy__:
            left = left._copy()

        if not isinstance(right, Term):
            right = Primitive(right)
        elif isinstance(right, QueryingProxy) and not right.__is_copy__:
            right = right._copy()

        left.right_parent = self
        right.left_parent = self

        self.left = left
        self.right = right
        self.operator = operator
        self.positions = {}
        super().__init__()


    def serialize(self, parentheses=True, full_names=True, for_header=False, translator=None) -> str:
        return self.operator

    def __str__(self):
        #not debug mode
        return self.str(parentheses=True, full_names=True)


    def str(self, parentheses=False, full_names=False, translator=None):
        res = ''        
        sorted_graph = self.sort_graph(translator)
        for term in sorted_graph:
            res += term.serialize(parentheses, full_names, translator=translator) + " "
        return res[:-1]



    def sort_graph(self, translator=None):
        res = []
        
        copies = self._simple_copy_terms(translator=translator)

        if self.debugger:
            self.debugger.show(copies)

        
        while len(copies):
            shifted = self.shift_graph(copies)
            if self.debugger and self.debugger.each_step:
                self.debugger.show(copies)
            term = self.terms_graph[shifted.id]
            res.append(term)
            del shifted

        return res



    def shift_graph(self, terms: List[Term]) -> Term:
        left_leaf = self.get_last_left_term(terms)

        if hasattr(left_leaf, 'right_parent') and left_leaf.right_parent\
            and hasattr(left_leaf, 'right') and left_leaf.right:
            left_leaf.right_parent.left = left_leaf.right
            left_leaf.right.right_parent = left_leaf.right_parent
            left_leaf.right.left_parent = None

        elif hasattr(left_leaf, 'right_parent') and left_leaf.right_parent:
            left_leaf.right_parent.left = None            

        elif hasattr(left_leaf, 'right') and left_leaf.right:
            left_leaf.right.left_parent = None            

        left_leaf_index = left_leaf.index_of(terms)
        del terms[left_leaf_index]
        return left_leaf

    def get_last_left_term(self, terms: List[Term]=None):
        if terms is None:
            terms = list(self.terms_graph.values())
        top = self.get_top(terms)

        res = top.go_until_hasattr('left_parent', 'left', terms)
        return res


    def get_last_right_term(self, terms: List[Term] = None):        
        if terms is None:
            terms = list(self.terms_graph.values())
        iterat = self.get_top(terms)

        res = iterat.go_until_hasattr('right_parent', 'right', terms)
        return res

    def get_top(self, terms: List[Term]=None) -> Term:
        if terms is None:
            terms = list(self.terms_graph.values())
        iterat = terms[0]

        res = iterat.go_until_hasattr('left_parent', 'right_parent', terms)
        return res
        

    def merge_graphs(self):
        terms_graph = {}
        left = self.left
        right = self.right

        # if left.__header__ or right.__header__:
        #     header = self.merge_headers(left.__header__, right.__header__)
        #     left.__header__ = header
        #     right.__header__ = header
        #     self.__header__ = header
            
        terms_graph.update(self.terms_graph)        
        terms_graph.update(left.terms_graph)        
        terms_graph.update(right.terms_graph)
        
        left.terms_graph = terms_graph
        right.terms_graph = terms_graph
        self.terms_graph = terms_graph

        return terms_graph

    def set_parentheses(self):
        self.parentheses = True
        left_leaf = self.left
        if isinstance(self.left, Binary):
            left_leaf = self.left.get_last_left_term()
        left_leaf.left_parentheses_count += 1

        right_leaf = self.right
        if isinstance(self.right, Binary):
            right_leaf = self.right.get_last_right_term()
        right_leaf.right_parentheses_count += 1


from typing import TYPE_CHECKING, Any, Union
if TYPE_CHECKING:
    from ..resource.db.table_column import TableColumn
    from ..resource.db.table import Table
    from ..resource.db.db import Db
from ..errors import DbError
from .term import Term, TermHeader
from ast import Name
from ast import Name, Constant
from ..errors import EpureParseError
from ..helpers.type_helper import check_type
from uuid import UUID

class Leaf(Term):
    left_parentheses_count = 0
    right_parentheses_count = 0

    def __init__(self) -> None:
        self.left_parentheses_count = 0
        self.right_parentheses_count = 0
        super().__init__()

    def append_parentheses(self, val):
        return '(' * self.left_parentheses_count + val + ')' * self.right_parentheses_count

    def __str__(self):
        return self.serialize(False, True)


class Primitive(Leaf, Constant):
    def __init__(self, val) -> None:
        self.val = val
        super().__init__()

    def serialize(self, parentheses=True, full_names=True, for_header=False, translator=None) -> str:
        res = ""
        if isinstance(self.val, list) or isinstance(self.val, tuple):
            res += "("
            for item in self.val:
                if isinstance(item, Leaf):
                    res += f'{item.serialize(parentheses, full_names, for_header)}, '
                    # temp.append(str(item))
                # elif isinstance(item,str):
                #     res += f"'{item}', "
                elif translator:
                    # temp.append(item.serialize(parentheses, full_names, for_header))
                    # res += f'{str(item)}, '
                    res += f'{translator(item, type(item))}, '
                elif isinstance(item,str):
                    res += f"'{item}', "
                else:
                    res += f'{str(item)}, '
                # else:
            res = res[:-2] + ")"
            # res += ")"
            # res = str(tuple(temp))
        elif translator:
            # res = str(self.val)
            res = translator(self.val, type(self.val))
        else:
            res = str(self.val)

        if not translator and (isinstance(self.val, str) or isinstance(self.val, UUID)):
            res = f"'{res}'"

        if not parentheses:
            return res
        res = self.append_parentheses(res)
        return res



class QueryingProxy(Leaf):
    if TYPE_CHECKING:
        __db__:Db
    __is_copy__ = False
    __qp_name__:str
    

    def _copy(self):
        raise NotImplementedError

    def in_header(self, header:Union[list,tuple]):
        raise NotImplementedError


class ColumnProxy(QueryingProxy, Name):
    if TYPE_CHECKING:
        __table__:Table
        __column__:TableColumn
        __table_proxy__ = None

    def __init__(self, db, table, column, table_proxy=None):
        self.__db__ = db
        self.__table__ = table
        self.__column__ = column
        self.__qp_name__ = self.serialize(parentheses=False, full_names=True)

        if table_proxy is None:
            table_proxy = TableProxy(db, table)
        self.__table_proxy__ = table_proxy
        super().__init__()



    def serialize(self, parentheses=True, full_names=True, for_header=False, translator=None) -> str:
        res = ''
        # if not full_names:
        #     res = self.__column__.name
        # el
        if for_header:
            res = self.__table__.header.serialize_read_column(self.__column__, full_names)
        elif not full_names:
            res = self.__column__.name
        else:
            table = self.__table__.full_name
            res = f'{table}.{self.__column__.name}'

        if not for_header and parentheses:
            res = self.append_parentheses(res)
        
        return res


    def _copy(self):
        res = ColumnProxy(self.__db__, self.__table__, self.__column__, self.__table_proxy__)
        res.__header__ = self.__header__
        res.__is_copy__ = True
        return res

    def in_header(self, header:Union[list,tuple]) -> bool:
        table_name = self.__table__.full_name
        for qp in header:
            check_type('qp', qp, [TableProxy, ColumnProxy])
            if isinstance(qp, ColumnProxy):
                if self.__qp_name__ == qp.__qp_name__:
                    return True
            if isinstance(qp, TableProxy):
                if qp.__qp_name__ == table_name:
                    return True
        return False


class TableProxy(QueryingProxy, Name):
    if TYPE_CHECKING:
        __table__:Table

    def __init__(self, db, table):
        self.__db__ = db
        self.__table__ = table
        self.__qp_name__ = self.serialize(parentheses=False, full_names=True)
        super().__init__()

    def __getattr__(self, attr_name: str) -> Any:
        if self.__is_copy__:
            raise AttributeError
        if attr_name not in self.__table__.header:
            raise AttributeError(f'column {attr_name} not in header of table {self.__table__.full_name}')
        column = self.__table__.header[attr_name]
        res = ColumnProxy(self.__db__, self.__table__, column, self)
        return res

    def __getitem__(self, *args) -> Any:
        res = TermHeader(list(*args))
        return res

    # def __call__(self, *args) -> Any:
    #     res = TermHeader(list(*args))
    #     return res

    def serialize(self, parentheses=True, full_names=True, for_header=False, translator=None) -> str:
        res = ''
        # if not full_names:
        #     res = self.__table__.name      
        # el
        if for_header:
            header = self.__table__.header
            for col in header:
                res += header.serialize_read_column(header[col], full_names) + ', '
            res = res[0:-2]
        else:
            res = f'{self.__table__.full_name}'

        if not for_header and parentheses:
            res = self.append_parentheses(res)
        
        return res

    def _copy(self):
        res = TableProxy(self.__db__, self.__table__)
        res.__header__ = self.__header__
        res.__is_copy__ = True
        return res

    def in_header(self, header:Union[list,tuple]) -> bool:
        for qp in header:
            check_type('qp', qp, [TableProxy, ColumnProxy])
            if isinstance(qp, TableProxy) and self.__qp_name__ == qp.__qp_name__:
                return True
        return False

class DbProxy(QueryingProxy):

    def __init__(self, db):
        self.__db__ = db
        super().__init__()

    def __getitem__(self, key:str):
        if key not in self.__db__:
            raise DbError(f'table {key} not in db {self.__db__.full_name}')
        res = TableProxy(self.__db__, self.__db__[key])
        return res

    def __iter__(self):
        return self.__db__.__iter__()

    def __len__(self):
        return self.__db__.__len__()


from typing import Any, List, Callable, Union


def check_type(var_name:str, value:Any, must_be:Union[type,List[type]]):
    err_text = f'{var_name} must be '
    check_all(err_text, value, must_be, isinstance)
    # for typ in must_be:
    #     if isinstance(value, typ):
    #         return
    #     else:
    #         err_text += f'{must_be} or'
    
    # raise TypeError(err_text[:-3])

def check_subclass(var_name:str, value:type, must_be:Union[type,List[type]]):
    err_text = f'{var_name} must subclass '
    check_all(err_text, value, must_be, issubclass)
    # for typ in must_be:
    #     if issubclass(value, typ):
    #         return
    #     else:
    #         err_text += f'{must_be} or'
    
    # raise TypeError(err_text[:-3])

def check_all(err_text:str, value:type, must_satisfy:Union[type,List[type]], check_func: Callable):
    if not isinstance(must_satisfy, list):
        must_satisfy = [must_satisfy]

    for must in must_satisfy:
        if check_func(value, must):
            return
        else:
            err_text += f'{must_satisfy} or'
    
    raise TypeError(err_text[:-3])


from typing import List, Dict, Union
from .term import Term, TermHeader
from .leaf import QueryingProxy
from ast import Constant, Name, BinOp, Eq, NotEq, LShift, RShift, BitXor, parse, dump, NodeVisitor, NodeTransformer, unparse, walk, iter_child_nodes, AST, Mod, And, GtE, Tuple
import re
# import astor
from ..helpers.string_helper import find_parentheses
from ..helpers.dict_helper import reverse_dict
from ..resource.db.table import Table
from ..helpers.type_helper import check_type
from uuid import uuid4
from ..errors import EpureParseError
from .leaf import Leaf

class JoinOperation:    
    table:str
    on_clause:str
    join_type:str
    join_id:str

    LEFT_TYPE='LEFT'
    RIGHT_TYPE='RIGHT'

    def __init__(self, table, on_clause, join_type='', join_id='') -> None:
        self.table = table
        self.on_clause = on_clause
        self.join_type = join_type
        self.join_id = join_id



class TermParser(NodeTransformer):
    header:List[QueryingProxy]
    joins:List[JoinOperation]
    parent = None
    resource:Table

    def __init__(self, resource:Table) -> None:
        self.resource = resource
        self.joins = []

    
    def parse(self, *args) -> str:
        #args
        header:Union[List[QueryingProxy], TermHeader] = None
        body:object = None
        full_names=True

        first = args[0]
        if isinstance(first, list) or isinstance(first, tuple):
            header = list(args[0])
            body = args[1]
            if len(args) > 2:
                full_names = args[2]
        else:
            header = []
            body = args[0]
            if len(args) > 1:
                full_names = args[1]

        check_type('body', body, [Term, str])
        if isinstance(header, TermHeader):
            header = header.val

        if isinstance(body, Term):
        # if isinstance(body, Term) and header:
            header = body.merge_headers(header, body.__header__)

        # if not header:
        #     raise EpureParseError('header not defined')

        body = body.str(True, full_names, self.resource.db.cast_py_db_val)
        self.joins.clear()
        # self.collect_joins(body)
        body = self.collect_joins(body)
        where_clause = self.remove_join_ids(body)

        if '@' in body:
            res = body.split(' @ ', 1)
            where_clause = res[1]
            header = res[0]
            if header[0] == '(':
                # header = eval(header)
                # header = header[1:-1]
                header = header.replace('(','')
                header = header.replace(')','')
                header = tuple(header.split(', '))
            else:
                header = (header,)
        
        if not header:
            header = (self.resource.querying_proxy,)
            # header = self.parse_for_headers(where_clause)
            # if header:
                

        res = self.resource.serialize_read(header, self.joins, where_clause, full_names)
        self.joins = []
        return res

    def parse_for_headers(self, body):
        res = []
        # if '@' in body:
        #    res = (body.split(' @ '))
        
        return res

    def collect_joins(self, body:str):
        tree = parse(body)
        self.visit(tree)
        res = unparse(tree)
        return res


    def remove_join_ids(self, body:str):

        for join in self.joins:
            body = body.replace(join.join_id, '')

        #needed?
        # match = re.search(pattern, repl)
        body = body.replace(' ^ ', '')
        body = body.replace('^ ', '')
        body = body.replace(' ^', '')
        body = body.replace('^', '')

        while '()' in body:
            body = body.replace('()', '')

        body = self.clear_parentheses(body, '( and')
        body = self.clear_parentheses(body, '(and')
        body = self.clear_parentheses(body, '( or')
        body = self.clear_parentheses(body, '(or')

        body = self.clear_parentheses(body, 'and )', 4)
        body = self.clear_parentheses(body, 'and)', 3)
        body = self.clear_parentheses(body, 'or )', 3)
        body = self.clear_parentheses(body, 'or)', 2)
        #needed?

        # body = body.replace('^', '')

        return body

    def clear_parentheses(self, body:str, pattern:str, offset:int=0):
        if pattern not in body:
            return body

            
        while pattern in body:
            first = body.index(pattern) + offset
            open_close_parentheses = find_parentheses(body)
            
            second = None
            if first in open_close_parentheses:
                second = open_close_parentheses[first]
            else:
                close_open_parentheses = reverse_dict(open_close_parentheses)
                second = close_open_parentheses[first]

            if second == None:
                raise EpureParseError()

            list_body = list(body)
            list_body[first] = ''
            list_body[second] = ''
            body = ''.join(list_body)

        return body


    def visit_BinOp(self, node:BinOp):
        self.generic_visit(node)
        op = node.op

        if isinstance(op,Mod) and isinstance(node.right.value, str):
            return Name(f"{unparse(node.left)} like {unparse(node.right)}")

        # if isinstance(op, And) and (isinstance() or )
        if not (isinstance(op, LShift) or isinstance(op, RShift)):
            return node

        join_type = JoinOperation.LEFT_TYPE if isinstance(op, LShift)\
            else JoinOperation.RIGHT_TYPE
        left = unparse(node.left)
        right = unparse(node.right)
        join_id = 'join_' + str(uuid4()).replace('-', '')

        join = JoinOperation(left, right, join_type, join_id)
        self.joins.append(join)
        return Name(join_id)
    
    # def serialize_tuple(self, ast_tuple:Tuple):
    #         res = []
    #         for const in ast_tuple.elts:
    #             res.append(const.value)
            
    #         return str(tuple(res))

    def visit_Compare(self, node:GtE):
        self.generic_visit(node)
        op = node.ops[0]

        if isinstance(op, GtE) and (isinstance(node.comparators[0], Tuple) or isinstance(node.comparators[0], BinOp)):
            # Ast_tuple_str = self.serialize_tuple(node.comparators[0])
            # return Name(f"{node.left.id} in {self.serialize_tuple(node.comparators[0])}")
            return Name(f"{unparse(node.left)} in {unparse(node.comparators[0])}")
        else:
            return node

    

    # def visit_LShift(self, node: LShift):        
    #     left = unparse(node.parent.left)
    #     right = unparse(node.parent.right)
    #     unparsed = unparse(node.parent)
    #     # unparsed = astor.to_source(node.parent)
    #     # unparsed = unparsed.replace('\n', '')
    #     # if unparsed[0] == '(' and unparsed[-1] == ')':
    #     #     unparsed = unparsed[1:-1]
    #     join = JoinOperation(left, right, JoinOperation.LEFT_TYPE, unparsed)
    #     self.joins.append(join)
    #     return node
    

    # def visit(self, node):
    #     node.parent = self.parent
    #     self.parent = node
    #     self.generic_visit(node)
    #     if isinstance(node, LShift):
    #         node = self.visit_LShift(node)
    #     if isinstance(node, AST):
    #         self.parent = node.parent
    #     return node


# from __future__ import annotations
# from ...errors import DbError

# class Term:

#     def __and__(self, other): #&
#         return self.binary_operation(other, 'and')

#     def __rand__(self, other): #&
#         return self.binary_operation(other, 'and', True)

#     def __or__(self, other:Term): #|
#         return self.binary_operation(other, 'or')

#     def __ror__(self, other:Term): #|
#         return self.binary_operation(other, 'or', True)

#     def __xor__(self, other:Term): #^
#         return self.binary_operation(other, '^')

#     def __rxor__(self, other:Term): #^
#         return self.binary_operation(other, '^', True)


#     def __lshift__(self, other:Term): #<<
#         return self.binary_operation(other, '<<')
#         # return self.join_operation(other, 'left')

#     def __matmul__(self, other:Term): #@
#         return self.binary_operation(other, '@')

#     def __rmatmul__(self, other:Term): #@
#         return self.binary_operation(other, '@', True)

#     def __eq__(self, other): #==
#         return self.binary_relation(other, '==')

#     def binary_relation(self, other:Term, operator:str):
#         if isinstance(self, PartialBinary):            
#             if not isinstance(other, Binary):
#                 return Binary(self, other, operator)
#             other.right = PartialBinary(other.left, other.right, other.operator)
#             other.left = self
#             other.operator = operator
#             return other
#         if isinstance(other, PartialBinary):            
#             other.left = PartialBinary(self, other.left, operator)
#             return other
#         return Binary(self, other, operator)

#     def binary_operation(self, other:Term, operator:str, right_self:bool = False):        
#         left = self
#         right = other
#         if right_self:
#             left = other
#             right = self
#         if isinstance(self, _PseudoColumn) or isinstance(other, _PseudoColumn) or not isinstance(other, Term):
#             return PartialBinary(left, right, operator)
#         if (isinstance(self, PartialBinary) and not isinstance(other, PartialBinary)) or\
#             isinstance(other, PartialBinary) and not isinstance(self, PartialBinary):
#             return PartialBinary(left, right, operator)
#         return Binary(left, right, operator)
        
#     # def join_operation(self, other:Term, join_type:str=''):
#     #     if not isinstance(self, _PseudoTable):
#     #         raise NotImplementedError(f'''join is cross-table relation, 
#     #                         its unkown for types: {type(self)}, {type(other)}''')
#     #     if isinstance(other, _PseudoColumn) or not isinstance(other, Term):
#     #         return PartialJoinBinary(self, other, join_type)
#     #     return JoinBinary(self, other, join_type)


# class Binary(Term):
#     left:Term
#     right:Term
#     operator:str    

#     def __init__(self, left, right, operator='') -> None:
#         self.left = left
#         self.right = right
#         self.operator = operator

#     def __str__(self) -> str:
#         return f'({self.left} {self.operator} {self.right})'


# class JoinBinary(Binary):    
#     table:str
#     on_clause:str
#     join_type:str
#     unparsed:str

#     LEFT_TYPE='LEFT'
#     RIGHT_TYPE='RIGHT'

#     def __init__(self, table, on_clause, join_type='', unparsed='') -> None:
#         self.table = table
#         self.on_clause = on_clause
#         self.join_type = join_type
#         self.unparsed = unparsed
        

# #     def __str__(self) -> str:
# #         return f'[{self.join_type} join on {self.left.__table__.full_name}:{self.right}]'

# # class PartialJoinBinary(JoinBinary):
# #     def __str__(self) -> str:
# #         return f'{self.join_type} join on {self.left.__table__.full_name}:{self.right}'

# class PartialBinary(Binary):
#     def __str__(self) -> str:
#         return f'{self.left} {self.operator} {self.right}'

# class Pseudo(Term):
#     pass
        


# class _PseudoColumn(Pseudo):
#     #overrided in pseudo.py
#     pass

# class _PseudoTable(Pseudo):
#     #overrided in pseudo.py
#     pass


# from .term import  Pseudo, _PseudoColumn, _PseudoTable, Term
# from .select_query import SelectQuery
# # Query, JoinClause, ,
# from ...errors import DbError
# from typing import Any, TYPE_CHECKING, cast
# from ...helpers.type_helper import check_type
# if TYPE_CHECKING:
#     from .table_column import TableColumn
#     from .table import Table
#     from .db import Db



# class PseudoColumn(_PseudoColumn):
#     if TYPE_CHECKING:
#         table:Table
#         column:TableColumn
#         db:Db
#     def __init__(self, db, table, column):
#         self.db = db
#         self.table = table
#         self.column = column

#     def __str__(self) -> str:
#         return f'{self.column.full_name}'
#         # return f'{self.table.full_name}.{self.column.full_name}'

# class PresudoTable(_PseudoTable):
#     if TYPE_CHECKING:
#         __table__:Table
#         __db__:Db
#     def __init__(self, db, table):
#         self.__db__ = db
#         self.__table__ = table
        

#     def __getattr__(self, attr_name: str) -> Any:
#         if attr_name not in self.__table__.header:
#             raise DbError(f'column {attr_name} not in header of table {self.__table__.full_name}')
#         column = self.__table__.header[attr_name]
#         res = PseudoColumn(self.__db__, self.__table__, column)
#         return res

#     def __str__(self) -> str:
#         return f'{self.__table__.full_name}'



# class PseudoDb(Pseudo):
#     if TYPE_CHECKING:
#         db:Db
#     def __init__(self, db):
#         self.db = db

#     def __getitem__(self, key:str):
#         if key not in self.db:
#             raise DbError(f'table {key} not in db {self.db.full_name}')
#         res = PresudoTable(self.db, self.db[key])
#         return res

#     def __iter__(self):
#         return self.db.__iter__()

#     def __len__(self):
#         return self.db.__len__()

#     def __call__(self, *args: list) -> Term:
#         select_header = args[0:-1]
#         select_body = args[-1]
#         check_type('select_body', select_body, Term)

#         res = SelectQuery(select_header, select_body)
        
#         table = list(self.db.tables.values())[0]
#         return table.serialize_read(res)


from __future__ import annotations
from typing import TYPE_CHECKING, Any, List
if TYPE_CHECKING:
    from .savable import Savable
from queue import Queue

# CREATE = 'CREATE'
# READ = 'READ'
# UPDATE = 'UPDATE'
# DELETE = 'DELETE'

class Resource():

    def create(self, savable:Savable, asynch:bool=False) -> object:
        raise NotImplementedError

    def serialize_for_create(self, savable:Savable, **kwargs) -> object:
        raise NotImplementedError

    def read(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def serialize_for_read(self, savable:Savable, **kwargs) -> object:
        raise NotImplementedError

    def update(self, savable:Savable, asynch:bool=False) -> object:
        raise NotImplementedError

    def serialize_for_update(self, savable:Savable,  **kwargs) -> object:
        raise NotImplementedError

    def delete(self, savable:Savable, asynch:bool=False):
        raise NotImplementedError

    def serialize_for_delete(self, savable:Savable,  **kwargs) -> object:
        raise NotImplementedError


    #all

    def create_all(self, savables:List[Savable]):
        raise NotImplementedError

    def update_all(self, savables:List[Savable], selector:object):
        raise NotImplementedError

    def delete_all(self, savables:List[Savable], selector:object):
        raise NotImplementedError



    def deserialize(self, savable:object, **kwargs) -> Savable:
        raise NotImplementedError

    def cache(self, script:str):
        raise NotImplementedError  

    def execute(self, script:str='') -> object:
        raise NotImplementedError

    def generate_id(self, savable:Savable=None):
        raise NotImplementedError


from __future__ import annotations
from types import NoneType
from typing import Any
from configparser import SectionProxy, RawConfigParser
from ...helpers.string_helper import is_float, is_int, is_bool

class IniSection:
    parser:RawConfigParser
    path:str

    def __init__(self, parser:RawConfigParser, path:str) -> None:
        self.parser = parser
        self.path = path

    def get_next(self, attr_name: str) -> IniSection:
        next_path = self.next_path(attr_name)
        if next_path in self.parser:
            return ActualIniSection(self.parser, next_path)
        else:
            return NoneIniSection(self.parser, next_path)
        
    def next_path(self, attr_name: str) -> str:
        return '.'.join([self.path, attr_name])
    

class NoneIniSection(IniSection):

    def __eq__(self, obj: Any) -> bool:
        if obj == None:
            return True
        return super().__eq__(obj)

    def __getattr__(self, attr_name: str) -> Any:
        return self.get_next(attr_name)

    def __bool__(self) -> bool:
        return False


class ActualIniSection(IniSection):

    section:SectionProxy

    def __init__(self, parser:RawConfigParser, path:str):
        if path not in parser:
            raise Exception('path not in parser')
        self.section = parser[path]
        super().__init__(parser, path)

    def __contains__(self, key: str) -> bool:
        return key in self.section


    def __getattr__(self, attr_name: str) -> Any:
        if attr_name not in self:
            return self.get_next(attr_name)

        attr_val = self.section[attr_name]
        return self._cast_val(attr_val)


    def _cast_val(self, attr_val: Any) -> Any:
        attr_is_float = is_float(attr_val)
        attr_is_int = is_int(attr_val, attr_is_float)

        if attr_is_int:
            return int(attr_val)
        elif attr_is_float:
            return float(attr_val)
        elif is_bool(attr_val):
            return attr_val.lower() == 'true'
        else:
            return attr_val






class EpureError(Exception):
    pass

class EpureParseError(Exception):
    pass

class UserInputError(EpureParseError):
    pass

class DbError(Exception):
    pass

class ResourceException(Exception):
    pass

class DefaultConstraintError(AttributeError):
    def get_message(self, class_name:str, field_name:str):
        return f'''field {field_name} of {class_name} doesn't have default value.
            Default type constraint (etc Default, NotNull, Prim)
                            must be provided with default value:
                            @epure()
                            class MyClass:
                                my_prim:Prim[int] = 0
                                ...
                            '''

    def __init__(self, class_name:str='', field_name:str='', message:str=''):
            if message:
                self.message = message
            else:
                self.message = self.get_message(class_name, field_name)
            super().__init__(self.message)

class MultipleInheritanceError(Exception):
    pass

class DeserializeError(Exception):
    pass


# from typing import List, Dict
# from .term import  Pseudo, JoinBinary, Term
# from .db_entity_resource import DbEntityResource
# from ast import Eq, NotEq, LShift, parse, dump, NodeVisitor, NodeTransformer, unparse, walk, iter_child_nodes, AST
# import re
# import astor
# from ...helpers.string_helper import find_parentheses


# class SelectQuery(NodeVisitor):
#     header:List[Pseudo]
#     joins:List[JoinBinary]
#     where_clause:str
#     parent = None

#     def __init__(self, header:List[Pseudo], body:Term) -> None:
#         self.header = header
#         self.joins = []
        
#         self.set_where_clause(str(body))


#     def set_where_clause(self, body:str):
#         tree = parse(body)
#         self.visit(tree)
#         body = self.remove_py_joins(body)        

#         self.where_clause = body


#     def remove_py_joins(self, body:str):

#         for join in self.joins:
#             body = body.replace(join.unparsed, '')

#         while '()' in body:
#             body = body.replace('()', '')

#         body = self.clear_parentheses(body, '( and')
#         body = self.clear_parentheses(body, '(and')
#         body = self.clear_parentheses(body, '( or')
#         body = self.clear_parentheses(body, '(or')
#         body = self.clear_parentheses(body, 'and )')
#         body = self.clear_parentheses(body, 'and)')
#         body = self.clear_parentheses(body, 'or )')
#         body = self.clear_parentheses(body, 'or)')

#         body = body.replace('^', '')

#         return body

#     def clear_parentheses(self, body:str, pattern:str):
#         if pattern not in body:
#             return body

            
#         while pattern in body:
#             first = body.index(pattern)
#             parentheses = find_parentheses(body)
#             second = parentheses[first]

#             list_body = list(body)
#             list_body[first] = ''
#             list_body[second] = ''
#             body = ''.join(list_body)

#         return body





#     def visit_LShift(self, node: LShift):        
#         left = unparse(node.parent.left)
#         right = unparse(node.parent.right)
#         unparsed = astor.to_source(node.parent)
#         unparsed = unparsed.replace('\n', '')
#         join = JoinBinary(left, right, JoinBinary.LEFT_TYPE, unparsed)
#         self.joins.append(join)
#         return node
    

#     def visit(self, node):
#         node.parent = self.parent
#         self.parent = node
#         self.generic_visit(node)
#         if isinstance(node, LShift):
#             node = self.visit_LShift(node)
#         if isinstance(node, AST):
#             self.parent = node.parent
#         return node


# from typing import Dict, Any
# from ...helpers.type_helper import check_type
# from .db_entity_resource import DbEntityResource
# from ...errors import DbError
# import inspect
# import sys


# class Pseudo:
#     pass


# class Query():

#     joins:list
#     condition:str
#     db:DbEntityResource
#     wrong_precedence:bool

#     def __init__(self, condition, joins = []):
#         self.condition = condition
#         self.joins = joins
#         # self.expr = f'{x}'

#     # def __rmatmul__(self, other:tuple):
#     def __xor__(self, other): #^
#         check_type('other', other, tuple)        
#         if isinstance(self, Query):
#             return SelectQuery(self.joins, self.condition, other)
#         raise NotImplementedError(f'''right join is cross-table relation, 
#                         its unkown for types: {type(self)}, {type(other)}''')

#     def get_brackets_term(self, other, operator:str, right:bool=False):
#         return '(' + self.get_term(other, operator, right) + ')'

#     def get_term(self, other, operator:str, right:bool=False):
#         if not isinstance(other, Query):
#             other = self.db.cast_py_db_val(type(other), other)
#         if right:
#             return str(other) + f' {operator} ' + str(self)
#         return str(self) + f' {operator} ' + str(other)


#     def logical_operation(self, other, operator:str, right:bool=False):

#         condition = self.get_condition(other, operator, right)

#         if hasattr(self, 'joins'):
#             joins = self.joins
#         else:
#             joins = []

#         if hasattr(other, 'joins'): 
#             joins = joins + other.joins

#         res = WhereClause(condition, joins)
#         if isinstance(self, Pseudo) or isinstance(other, Pseudo):            
#             res.wrong_precedence = True

#         return res
           
#     def __and__(self, other): #&
#         return self.logical_operation(other, 'AND')


#     def __or__(self, other): #|
#         return self.logical_operation(other, 'OR')

#     def __rand__(self, other): #&
#         return self.logical_operation(other, 'AND', True)


#     def __ror__(self, other): #|
#         return self.logical_operation(other, 'OR', True)


#     def __lshift__(self, other): #<<
#         raise NotImplementedError(f'''left join is cross-table relation, 
#                         its unkown for types: {type(self)}, {type(other)}''')

#     def __rshift__(self, other): #>>
#         raise NotImplementedError(f'''right join is cross-table relation, 
#                         its unkown for types: {type(self)}, {type(other)}''')

#     def get_condition(self, other, operator, right:bool=False):
#         if (hasattr(self, 'wrong_precedence') and self.wrong_precedence) or \
#             (hasattr(other, 'wrong_precedence') and other.wrong_precedence) or\
#                 isinstance(self, Pseudo) or isinstance(other, Pseudo):

#             return self.get_term(other, operator, right)
        
#         return self.get_brackets_term(other, operator, right)

#     def __eq__(self, other):
#         operator = '='
        
#         if other is None:
#             operator = 'is'

#         condition = self.get_condition(other, operator)

#         return WhereClause(condition, [])


#     def __ne__(self, other): #!=
#         if other is None:
#             return self.get_brackets_term(other, ' is not ')
#         return self.get_brackets_term(other, '!=')


#     def __lt__(self, other): #<
#         return self.get_brackets_term(other, '<')

#     def __le__(self, other): #<=
#         return self.get_brackets_term(other, '<=')
#     def __gt__(self, other): #>
#         return self.get_brackets_term(other, '>')
#     def __ge__(self, other): #>=
#         return self.get_brackets_term(other, '>=')


#     # def __invert__(self): #~
#     #     expr = '( invert ' + self.expr + ')'
#     #     return Query(expr)

#     def __str__(self) -> str:
#         return self.condition


# class SelectQuery(Query):
#     selected:Any

#     def __init__(self, condition, joins, selected):
#         self.selected = selected
#         return super().__init__(condition, joins)

# class JoinClause(Query):
#     direction:str
#     table:Any
#     operator:str

#     def __init__(self, condition, direction, table, operator:str=''):
#         if operator:
#             self.operator = operator
#         self.direction = direction
#         self.table = table
#         return super.__init__(condition)


#     def logical_operation(self, other, operator: str):
#         if isinstance(other, JoinClause):
#             other.operator = operator
#             return WhereClause('', [self, other])
#         if isinstance(other, WhereClause):
#             self.operator = operator
#             return WhereClause(other.condition, other.joins + [self])
#         raise NotImplementedError(f'''operation {operator} unkown for types: 
#             {type(self)}, {type(other)}''')
        

# class WhereClause(Query):

#     def logical_operation(self, other, operator: str):
#         if isinstance(other, JoinClause):
#             other.operator = operator
#             return WhereClause(self.condition, self.joins + [other])
#         if isinstance(other, WhereClause):
#             return super().logical_operation(other, operator)
#         raise NotImplementedError(f'''operation {operator} unkown for types: 
#             {type(self)}, {type(other)}''')


from __future__ import annotations
from abc import abstractmethod
from typing import Dict, Any, Type, cast, List
import logging

from .db_entity import DbEntity

from ..savable import Savable
from .table import Table
from .table_storage import TableStorage
from urllib.parse import urlparse
from ...errors import EpureError
from ...helpers.type_helper import check_type



class Db(TableStorage):

    database:str
    user:str
    password:str
    host:str
    port:str    
    params:Dict[str,str]

    log_level:int = logging.NOTSET
    logger:logging.Logger
    connection:Any
    cache_queue:List[str] #Queue
    
    


    @abstractmethod
    def __init__(self, 
                connect_str:str='', 
                database:str='', 
                user:str='', 
                password:str='',
                host:str='', 
                port:str='', 
                default_namespace='', 
                log_level:int = logging.NOTSET,                
                default_table_type:Type[Table]=None,
                migrate_on_delete:bool=False):
        


        connect_params = urlparse(connect_str)

        self.database = (database or connect_params.scheme)
        self.user = user or str(connect_params.username)
        self.password = password or str(connect_params.password)
        self.host = host or str(connect_params.hostname)
        self.port = port or str(connect_params.port)        

        if default_namespace:
            self.default_namespace = default_namespace

        if log_level:
            self.log_level = log_level

        self.cache_queue = []
        self.namespaces = []
        self.tables = {}
        self.set_logger()

        return super().__init__(default_table_type, migrate_on_delete)



    def create(self, db_entity: Savable) -> DbEntity:
        check_type('db_entity', db_entity, DbEntity)

        if not db_entity.namespace in self.namespaces:
            self.create_namespace(db_entity.namespace)

        if isinstance(db_entity, Table):
            return self.create_table(db_entity)
        raise NotImplementedError(f'createion not implemented for type {type(db_entity)}')


    def update(self, db_entity: Savable) -> DbEntity:
        check_type('db_entity', db_entity, DbEntity)
        if isinstance(db_entity, Table):
            return self.update_table(db_entity)
        raise NotImplementedError(f'update not implemented for type {type(db_entity)}')


    def execute(self, script: str = '') -> list:
        result = []
        
        if hasattr(self, 'logger') and self.logger:
            self.logger.debug(script)
        try:
            script = self.get_cache() + script
            script = script.replace("\n", "")
            result = self._execute(script)
        except Exception as ex:
            err = EpureError(f'unnable execute script {script}: {ex}') \
                .with_traceback(ex.__traceback__)
            if self.logger:
                self.logger.error(err, exc_info=True)
                
            raise err
        return result

    def _execute(self, script: str = '') -> list:
        raise NotImplementedError()

    def get_cache(self) -> str:
        res = self.cache_queue
        if not res:
            return ''
        res = '\n'.join(res)
        res = res + '\n'
        self.cache_queue = []
        return res

    def set_logger(self) -> Any:
        if self.log_level > logging.NOTSET:
            self.logger = logging.getLogger(__name__)
            
            self.logger.setLevel(self.log_level)
            fileHandler = logging.FileHandler('epure_db.log', mode='a+')
            fileHandler.setLevel(self.log_level)
            formater = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s:\n %(message)s')
            fileHandler.setFormatter(formater)

            self.logger.addHandler(fileHandler)


from types import NoneType
from .resource import Resource
from typing import Any, Dict, get_type_hints

class Savable(Resource):    
    
    resource: Resource
    is_saved:bool=False
    _annotations: Dict[str,Any]
    #append to it, if u want excude
    __exclude__:list = ['resource', 'is_saved', '_annotations', 'cache_queue']
    

    def __init__(self, resource:Resource=None) -> None:
        if resource != None:
            self.resource = resource
        super().__init__()

    @property
    def annotations(self) -> Dict[str,Any]:
        if not hasattr(self, '_annotations'):
            if isinstance(self, type):                
                self._annotations = get_type_hints(self)
            else:
                self.__class__._annotations = get_type_hints(self.__class__)
        return self._annotations

    def save(self, asynch:bool=False):
        pass

    def to_json(self):
        raise NotImplementedError

    @classmethod
    def is_excluded(cls, atr_name:str, type_hint:Any='') -> bool:
        if hasattr(cls, '__exclude__') and atr_name in cls.__exclude__:
            return True
        if atr_name[:2] == "__" and atr_name[-2:] == "__":
            return True
        if atr_name in ("prepared_resource"):
            return True
        if type_hint in (NoneType, None):
            return True
        return False

    def execute(self, script: str = '') -> object:
        return self.resource.execute(script)


from __future__ import annotations
from types import NoneType
from typing import TYPE_CHECKING, Dict, Union, List, ItemsView, Any, Type, Callable, cast, Optional
from ..savable import Savable
from .constraint import Constraint
from .db_entity_resource import DbEntityResource
import re
from ...named import Named


class TableColumn(Savable, Named):
    py_type:type
    
    @property
    def is_deleted(self):
        # res = re.match(r'.*deleted_[a-f0-9]{32}', self.name)
        res = self.name[-10:-4] == '___del'
        if res:
            return res
        return res

    def __init__(self, name:str, py_type:type=NoneType) -> None:
        self.py_type = py_type
        self.name = name
        super().__init__()

    
    def __eq__(self, other:TableColumn) -> bool:
        if self.name != other.name:
            return False            
        return self.py_type == other.py_type

    def serialize_type(self, db:DbEntityResource):
        py_type = self.py_type
        if isinstance(py_type, Constraint):
            return db.serialize_constraint(py_type)
        return db.get_db_type(py_type)


from ..db.table_header import TableHeader
from ..db.table_column import TableColumn
from ..db.db_entity_resource import DbEntityResource
import re
from ...helpers.string_helper import is_float


class GresHeader(TableHeader):
    
    def _deserialize(self, column_dict: dict) -> dict:
        column_name = column_dict['column_name']
        db_type = column_dict['data_type']
        not_null = (column_dict['is_nullable'] == 'NO')
        default = column_dict['column_default']
        uniq = (column_dict['constraint_type'] == 'UNIQUE')
        prim = (column_dict['constraint_type'] == 'PRIMARY KEY')
        foreign = (column_dict['constraint_type'] == 'FOREIGN KEY')

        if default:
            default = self.cast_db_py_column_default(default, db_type)

        res = {'column_name': column_name,
                'not_null': not_null,
                'db_type': db_type,
                'default':default,
                'uniq':uniq,
                'prim':prim,
                'foreign':foreign}

        if foreign:
            foreign_schema = column_dict['foreign_schema']
            foreign_table = column_dict['foreign_table']
            foreign_table = foreign_schema + '.' + foreign_table
            foreign_column = column_dict['foreign_column']

            res.update({'foreign_schema':foreign_schema,
                        'foreign_table':foreign_table,
                        'foreign_column':foreign_column})
                                
        return res
        

    def cast_db_py_column_default(self, default:str, db_type:str):
        res = re.sub('::[a-z]*', '', default)
        if res[0] == "'" and res[-1] == "'":
            res = res[1:-1]

        elif is_float(res):
            res = float(res)
        elif res[:5] == 'point':
            pattern = re.compile(r"point\(\((.*)\).*\((.*)\).*\)")
            res = pattern.findall(res)[0]
            res = complex(float(res[0]), float(res[1]))

        return res

    def _rename_column_script(self, table_name:str, old_name:str, new_name:str) -> str:
        return f'''
            ALTER TABLE {table_name} RENAME {old_name} TO {new_name};
        '''
        
    
    def _set_not_null_script(self, table_name:str, column_name:str) -> str:
        return f'''
            ALTER TABLE {table_name} ALTER COLUMN {column_name} SET NOT NULL;
        '''

    def _set_nullable_script(self, table_name:str, column_name:str) -> str:
        return f'''
            ALTER TABLE {table_name} ALTER COLUMN {column_name} DROP NOT NULL;
        '''

    def _create_column_script(self, table_name:str, column:TableColumn, db:DbEntityResource) -> str:        
        db_type = column.serialize_type(db) #db.get_db_type(column.column_type)
        
        create_script = f'ALTER TABLE {table_name} ADD COLUMN {column.name} {db_type};'
        return create_script

    def _migrate_columns_script(self, table_name:str, from_column:str, to_column:str) -> str:
        return f'''
            UPDATE TABLE {table_name} SET {to_column} = {from_column};
        '''
    
    def serialize_read_column(self, column:TableColumn|str, full_names:bool=True):
        
        column_name = column.name if isinstance(column,TableColumn) else column

        if not full_names: 
            return column_name        
        
        if isinstance(column,TableColumn):
            table_name = self.table.full_name
            column_name = f'{table_name}.{column_name}'

        alias = column_name.replace('.', '___')
        column_name = f'{column_name} as {alias}'
        return column_name


from ..resource import Resource
from typing import Dict, List, Any, Callable
from .constraint import Constraint
from ...errors import DbError
from ..savable import Savable
from uuid import UUID


class DbEntityResource(Resource):

    py_db_types:Dict[type, str]
    db_py_types:Dict[str, type]
    default_namespace:str
    py_db_operators:Dict[str, str] = {
        '==': '=',
    }
    json_serializer:Savable
    # py_db_key_words:Dict[str, str] = {
    #     'where': 'where'
    # }
    

    def serialize_constraint(self, constraint:Constraint) -> str:
        raise NotImplementedError

    def get_default_namespace(self):
        return self.default_namespace

    def get_db_type(self, py_type:type):
        if py_type in self.py_db_types:
            return self.py_db_types[py_type]
        if Any not in self.py_db_types:
            raise DbError(f'''type {py_type} unknown for 
                {self.full_name} specify Any db type for using unknown types''')
        return self.py_db_types[Any]

    def get_py_type(self, db_type:str):
        return self.db_py_types[db_type]

    def same_db_type(self, first:type, second:type) -> bool:
        if first == second:
            return True
        res = self.get_db_type(first) == self.get_db_type(second)
        return res

    def cast_py_db_val(self, val:Any, py_type:type) -> str:
        raise NotImplementedError


    def cast_db_py_val(self, val:Any, py_type:type) -> Any:
        # if type(val) == Decimal:
        if val is None:
            return val
        if py_type in (int, float, bool, UUID):
            return py_type(val)
        if self.same_db_type(py_type, Any):
            return self.json_serializer.deserialize(val)
        # if py_type == UUID:
            # return str(val)
        return val
        # return py_type(val) instead of 54,55.




def screen_regex_specials(val:str):
    res = ''
    specials = ('.', '+', '*', '?', '^', '$', '(', ')', '[', ']', '{', '}', '|', '\\', '<', '>')
    for char in val:
        if char in specials:
            res += f"\\{char}"
            continue
        res += char
    return res

#return dict: {open bracket: close bracket}
def find_parentheses(val:str) -> dict:
    res = {}
    pstack = []

    for i, c in enumerate(val):
        if c == '(':
            pstack.append(i)
        elif c == ')':
            if len(pstack) == 0:
                raise IndexError("No matching closing parens at: " + str(i))
            res[pstack.pop()] = i

    if len(pstack) > 0:
        raise IndexError("No matching opening parens at: " + str(pstack.pop()))

    return res


def is_float(val: str) -> bool:
    if val is None:
            return False
    first_char = val[0]
    if not (first_char == '-' or first_char.isdecimal()):
        return False

    other_chars = val[1:]
    if other_chars == '':
        return True
    return other_chars.replace('.', '', 1).isdecimal()

def is_int(val: str, val_is_float:bool=None) -> bool:
    if val_is_float is None:
        val_is_float = is_float(val)
   
    return (val_is_float and val.count('.') == 0)

def is_bool(val:str) -> bool:
    return val.lower() in ('true', 'false')


from uuid import UUID
from ..db.db import Db
import psycopg2
import psycopg2.extras
from typing import cast, Type, Dict, Any, Callable, List, Tuple
from types import NoneType, LambdaType
from itertools import groupby
from .gres_table import GresTable
from ..db.table import Table
from datetime import timedelta, datetime
from ipaddress import _IPAddressBase
from ...named import FullName, SnakeCaseNamed
from ..savable import Savable
import logging
from inflection import underscore
from ...helpers.type_helper import check_type
from ..db.constraint import Constraint, Default, Foreign, Prim, NotNull, Uniq
from ...errors import DbError
from ..file.json_file import JsonFile
from .jsonb_table import JsonbTable
from decimal import Decimal

class GresDb(Db):

    default_table_type: Type[Table] = GresTable
    nosql_table_type: Type[Table] = JsonbTable
    default_namespace:str='Public'
    log_level:int = logging.NOTSET


    def __init__(self, connect_str:str='', database:str='', user:str='', password:str='',
             host:str='', port:str='', default_namespace='', log_level:int = logging.NOTSET,
             migrate_on_delete:bool=False):

        super().__init__(connect_str, 
            database=database, 
            user=user, 
            password=password, 
            host=host, 
            port=port,
            default_namespace=default_namespace,
            log_level=log_level)

        self.params = {
            'database': self.database, 
            'user': self.user, 
            'password': self.password, 
            'host': self.host, 
            'port': self.port
        }

        print(self.execute("SELECT 'test request'"))

        self._set_namespaces()
        
        self._set_tables()

        self.json_serializer = JsonFile('')


    def read(self, selector:object=None, **kwargs) -> Any:
        
        table_name = str(selector)
        full_table_name = self._get_full_table_name(table_name)


        collumns = self.execute(f'''SELECT cols.table_schema, cols.table_name, 
                    cols.column_name, cols.is_nullable, cols.data_type, cols.column_default,
                    cols_constr.table_schema AS foreign_schema,
				    cols_constr.table_name AS foreign_table,
				    cols_constr.column_name AS foreign_column,
                    constr.constraint_type
                    FROM information_schema.columns cols
                    left join information_schema.constraint_column_usage cols_constr
                    	on cols.table_schema = cols_constr.table_schema and cols.table_name = cols_constr.table_name
                    	and cols.column_name = cols_constr.column_name
                    left join information_schema.table_constraints constr
                    	on cols_constr.constraint_name = constr.constraint_name WHERE
                    cols.table_schema = \'{full_table_name.namespace}\' AND
                    cols.table_name = \'{full_table_name.name}\' ''')

        if len(collumns) > 0:
            table = self.deserialize_table(collumns)
            self._set_table(table)
            return table
        return []
    

    def serialize_namespace_for_create(self, namespace) -> object:
        return f'''CREATE SCHEMA {namespace};'''


    def serialize_for_create(self, table: Savable, **kwargs) -> object:
        check_type('table', table, self.default_table_type)
        table = cast(Table, table)

        header = table.serialize_header(self)
        column_defenitions = []
        for column in header[:-1]:
            column_defenitions.append(
                column['column_name'] + " " + column['column_type'] + ","
            )

        last_column = header[len(header)-1]
        column_defenitions.append(
            last_column['column_name'] + " " + last_column['column_type']
        )

        script = " \n ".join(column_defenitions)
        
        script = f'''
        CREATE TABLE {table.full_name} (
            {script}
        );'''

        return script



    def _execute(self, script: str = '') -> list:
        result = []
        with psycopg2.connect(**self.params) as connection:
            with connection.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
                cursor.execute(script)
                if cursor.rowcount > 0:
                    try:
                        result = cursor.fetchall()
                    except psycopg2.ProgrammingError as err:
                        self.logger.info(err)

        return result




    def _deserialize_table_name(self, table_columns:list) -> FullName:
        first_columns = table_columns[0]

        table_name = first_columns['table_name']
        namespace = first_columns['table_schema']
        return SnakeCaseNamed(name=table_name, namespace=namespace)

    def get_default_namespace(self):
        return underscore(self.default_namespace)
      
    def _set_namespaces(self):
        script = '''
          SELECT schema_name
                    FROM information_schema.schemata
        '''
        all_namespaces = self.execute(script)

        # for namespace in all_namespaces:
        #     self.namespaces.append(namespace[0])

        self.namespaces = [row[0] for row in all_namespaces]

    def _set_tables(self):
        script = '''
          SELECT cols.table_schema, cols.table_name, 
                    cols.column_name, cols.is_nullable, cols.data_type, cols.column_default,
                    cols_constr.table_schema AS foreign_schema,
				    cols_constr.table_name AS foreign_table,
				    cols_constr.column_name AS foreign_column,
                    constr.constraint_type
                    FROM information_schema.columns cols
                    left join information_schema.constraint_column_usage cols_constr
                    	on cols.table_schema = cols_constr.table_schema and cols.table_name = cols_constr.table_name
                    	and cols.column_name = cols_constr.column_name
                    left join information_schema.table_constraints constr
                    	on cols_constr.constraint_name = constr.constraint_name
                    ORDER BY cols.table_schema, cols.table_name
        '''
        all_collumns = self.execute(script)    
        
        for key, group in groupby(all_collumns,  lambda row: row['table_schema'] + '.' + row['table_name']):
            group = list(group)
            # if group[0]['table_name'] == 'default_epure':
            #     breakpoint()
            table = self.deserialize_table(group)
            self._set_table(table)

    def _get_table_type(self, table_columns:list):
        jsonb_cols = list(filter(lambda c: c['column_name'] == 'jsonb___data', table_columns))
        if len(jsonb_cols) > 0:
            return JsonbTable
        return self.default_table_type
        # if len(table_columns) > 2:
        #     return self.default_table_type
        # jsonb_cols = list(filter(lambda c: c['data_type'] == 'jsonb', table_columns))
        # if len(jsonb_cols) != 1:
        #     return self.default_table_type
        # return JsonbTable

    def serialize_constraint(self, constraint:Constraint) -> str:
        
        db_type = self.get_db_type(constraint.py_type)
        _default = getattr(constraint, 'default', '')
        default = ''
        if _default:
            default = self.cast_py_db_val(_default, constraint.py_type)
        
        origin = constraint.__origin__
        
        if origin == Default:
            return f"{db_type} DEFAULT {default}"
        elif origin == NotNull:
            return f"{db_type} NOT NULL DEFAULT {default}"
        elif origin == Uniq:
            return f"{db_type} UNIQUE"
        elif origin == Prim:
            return f"{db_type} PRIMARY KEY DEFAULT {default}"
        elif origin == Foreign:
            return f"{db_type}" #Foreign not realy implemented yet, because of ciclyc cases
                
        raise DbError('unknown constraint')
        
    def cast_py_db_val(self, val:Any, py_type:type) -> str:
        if val == None:
            return 'NULL'
        if py_type in (int, float, bool):
            return str(val)
        if py_type in (str, UUID):
            res = f"'{str(val)}'"
            return res
        if py_type in (bytes, bytearray):        
            val = val.decode()
            val = val.replace("\x00","")
            res = f"'{val}'"
            return res
        if py_type == complex:            
            return f"point({val.real}, {val.imag})"

        json = self.json_serializer.serialize_for_update(val)
        return f"'{json}'"

        
    py_db_types:Dict[type, str] = {
        str: 'text',
        int: 'bigint',
        float: 'numeric',        
        complex: 'point',
        UUID: 'uuid',
        bool: 'boolean',
        Any: 'jsonb',

        #bytea siblings
        bytes: 'bytea',
        bytearray: 'bytea',

        #json siblings
        range: 'json',
        list: 'json',
        List: 'json',
        tuple: 'json',
        Tuple: 'json',
        object: 'json',
        dict: 'json',
        Dict: 'json',
        Dict[int,str]: 'json',
        set: 'json',
        frozenset: 'json',
        memoryview: 'json',
        LambdaType: 'json',
        Callable: 'json'
    }

    db_py_types:Dict[str, type] = {
        'bigint': int, #-9223372036854775808 to +9223372036854775807
        'text': str,
        # 'json': object,
        'name': str,
        'ARRAY': list,
        'oid': int, #0 to 4294967295
        'interval': timedelta,
        'smallint': int, #-32768 to +32767
        'inet': _IPAddressBase, #IPv4Address or IPv6Address ipaddress.ip_address        
        'pg_node_tree': str, #representation of parsed sql query
        'boolean': bool,
        'numeric': float, #up to 131072 digits before the decimal point; up to 16383 digits after the decimal point
        'decimal': float, #up to 131072 digits before the decimal point; up to 16383 digits after the decimal point        
        'anyarray': list,
        'regproc': str, #sql procidure name
        'regtype': str, #sql type name
        'timestamp' : datetime,
        'timestamp with time zone': datetime,
        'time' : datetime,
        'time with time zone' : datetime,
        'double precision': float,
        'real': float,
        '"char"': str,
        'character varying': str,
        'character': str,
        'point' : complex,
        'json': object,
        'jsonb': Any,
        'uuid' : UUID,
        'bytea': bytes,

        'pg_ndistinct': int,
        'pg_mcv_list': int,
        'pg_dependencies': int,        
        'xid': int,        
        'integer': int,
        'pg_lsn': int,        
        'date': int,
        'smallserial':int,
        'serial': int,
        'bigserial': int,
        'bit': int,
        'bit varying': int,
        'box': int,
        'cidr': int,
        'circle': int,
        'line' : int,
        'lseg' : int,
        'macaddr' : int,
        'macaddr8' : int,
        'money' : int,
        'path' : int,
        'pg_snapshot' : int,        
        'polygon' : int,
        'tsquery' : int,
        'tsvector' : int,
        'txid_snapshot' : int,        
        'xml' : int
    }


from __future__ import annotations
from .node import Node, TableNode
from typing import Any, Dict, get_type_hints, Tuple
from ..savable import Savable
from ...errors import MultipleInheritanceError

class ProtoMeta(type):
    # def proto_cls(cls):
    #     if '_proto_cls' in vars(cls).keys():
    #         return cls._proto_cls

    @classmethod
    def savable_classes(mcl, bases):
        res = list(filter(lambda base: issubclass(base, Savable), bases))
        return res

    def __new__(mcl, name:str, bases:Tuple[type], namespace: dict[str, Any]) -> ProtoMeta:
        savable_classes = mcl.savable_classes(bases)
        if len(savable_classes) > 1:
            raise MultipleInheritanceError(f'{name} has more than one Savable parent: {str(savable_classes)}')
        res = super().__new__(mcl, name, bases, namespace)
        if len(savable_classes) == 1:
            res.__proto_cls__ = savable_classes[0]
        return res

class Proto(ProtoMeta, TableNode):
    @property
    def annotations(self) -> Dict[str,Any]:
        if not hasattr(self, '_annotations'):
            if isinstance(self, type):
                self._annotations = self.get_self_annotations()
            else:
                self.__class__._annotations = self.get_self_annotations()
        return self._annotations

    @property
    def all_annotations(self) -> Dict[str,Any]:
        if not hasattr(self, 'all_annotations'):
            if isinstance(self, type):                
                self._all_annotations = get_type_hints(self)
            else:
                self.__class__._all_annotations = get_type_hints(self.__class__)
        return self._all_annotations

    @classmethod
    def get_self_annotations(cls):
        res = []
        all_annotations = get_type_hints(cls)
        self_annotations  = cls.__annotations__
        for annot_name in self_annotations:
            res[annot_name] = all_annotations[annot_name]
        res['__proto__'] = cls.__bases__[0]
        return res


    @property
    def __proto__(self) -> Node:
        proto_obj = self.__proto_cls__.__new__(proto_obj)
        for key in proto_obj.annotations:
            if key in vars(self):
                val = getattr(self, key)
                setattr(proto_obj, val)
        return proto_obj


    def set_proto_fields(self, proto:Savable):
        for attr_name in vars(proto):
            if attr_name in self.all_annotations:
                val = getattr(proto, attr_name)
                setattr(self, val)

    @classmethod
    def is_excluded(cls, atr_name:str, type_hint:Any='') -> bool:
        if atr_name == '__proto__':
            return False
        return super().is_excluded(atr_name, type_hint)


import networkx as nx
import matplotlib.pyplot as plt
from uuid import UUID, uuid4

class TermDebugger:
    each_step = False

    def __init__(self, each_step=False) -> None:
        self.each_step = each_step

class MatplotTermDebugger(TermDebugger):
    debug_tree_step = 10
    # positions = {}

    def __init__(self, debug_tree_step=None) -> None:
        if debug_tree_step:
            self.debug_tree_step = debug_tree_step
        super().__init__()

    def show(self, terms):

        nx_edges = []
        for term in terms:
            term_name = str(term.id)[0:4]
            term_name = f'{term.val}_{term_name}'
            if term.parentheses:
                term_name = term_name + '_()'
            term.name = term_name


            if hasattr(term, 'left') and term.left:
                left_name = str(term.left.id)[0:4]
                left_name = f'{term.left.val}_{left_name}'
                if term.left.parentheses:
                    left_name = left_name + '_()'
                nx_edges.append((term_name, left_name))


            if hasattr(term, 'right') and term.right:
                right_name = str(term.right.id)[0:4]
                right_name = f'{term.right.val}_{right_name}'
                if term.right.parentheses:
                    right_name = right_name + '_()'
                nx_edges.append((term_name, right_name))


        positions = self.get_positions(terms)

        nx_graph = nx.DiGraph()
        nx_graph.add_edges_from(nx_edges)
        
        plt.figure(str(uuid4()))
        nx.draw_networkx(nx_graph, pos=positions)
        plt.show()


    def get_positions(self, terms):
        tops = self.get_tops(terms)
        res = {}
        cur_pos = (0, 0)
        for top in reversed(tops):
            positions = self.get_positions_recurce(top, terms, cur_pos)
            cur_pos = self._get_right_up_position(positions)
            cur_pos = (cur_pos[0] + self.debug_tree_step, cur_pos[1])
            res.update(positions)
        return res

    def get_positions_recurce(self, term, terms, cur_pos):
        res = {}
        res[term.name] = cur_pos
        if hasattr(term, 'left') and term.left:
            left_position = (cur_pos[0] - 1, cur_pos[1] - 1)
            left_positions = self.get_positions_recurce(term.left, terms, left_position)
            res.update(left_positions)
        if hasattr(term, 'right') and term.right:
            right_position = (cur_pos[0] + 1, cur_pos[1] - 1)
            right_positions = self.get_positions_recurce(term.right, terms, right_position)
            res.update(right_positions)
        return res

    def get_tops(self, terms):
        not_tops = []
        for term in terms:
            if hasattr(term, 'left_parent') and term.left_parent:
                not_tops.append(term)
            elif hasattr(term, 'right_parent') and term.right_parent:
                not_tops.append(term)
        tops = []
        for term in terms:
            if term.index_of(not_tops) == None:
                tops.append(term)
        return tops

    def _get_right_up_position(self, positions):
        if not positions:
            return (0, 0)
        ziped = list(zip(*positions.values()))
        xes = ziped[0]
        yes = ziped[1]
        return (max(xes), max(yes))


from ..db.table import Table
from .gres_header import GresHeader
from .gres_entity import GresEntity
from typing import Dict, Any, List
from ..savable import Savable
from ...errors import  DbError, EpureParseError
# from ..db.select_query import SelectQuery
# from ..db.term import JoinBinary, Pseudo, _PseudoColumn, _PseudoTable
from ...parser.leaf import QueryingProxy, ColumnProxy, TableProxy
from ...parser.term_parser import JoinOperation
from uuid import uuid4

class GresTable(Table, GresEntity):
    def _set_header(self, header):
        if header == None:
            header = GresHeader(table=self)
        
        if isinstance(header, Dict):
            header = GresHeader(columns=header, table=self)
        
        self.header = header
        self.header.resource = self



    def serialize_for_create(self, node: Savable, **kwargs) -> object:

        node_dict = self._serialize(node)

        columns = ""
        values = ""

        for name, val in node_dict.items():
            columns = columns + name + ', '
            values = values + val + ', '
        columns = columns[:-2]
        values = values[:-2]

        res = f'INSERT INTO {self.full_name}({columns}) VALUES ({values}) returning node_id;'
        return res



    def serialize_read(self, header, joins, where_clause, full_names) -> str:
        header = self._add_node_id_fields(header)
        res_header = self.serialize_read_header(header, full_names)
        res_joins = self.serialize_joins(joins)        

        res = f'{res_header} \n {res_joins} WHERE \n {where_clause}'
        res = self.replace_operators(res)
        return res



    def serialize_read_header(self, header:List[QueryingProxy], full_names:bool):
        res = 'SELECT'
        for item in header:
            if isinstance(item, ColumnProxy):
                res += f' {item.serialize(False, full_names, True)},'
            elif isinstance(item, TableProxy):
                res += f' {item.serialize(False, full_names, True)},'
            elif isinstance(item, str):
                sp = item.split('.')
                if len(sp) == 2:
                    proxy = self.db[item].querying_proxy
                    res += f' {proxy.serialize(False, full_names, True)},'
                else:
                    res += f' {self.header.serialize_read_column(item, full_names)},'
            else:
                raise EpureParseError('select header item must be ColumnProxy or TableProxy')
        res = res[:-1]

        first_item = header[0]
        table_name = ''
        if isinstance(first_item, ColumnProxy):
            table_name = first_item.__table__.full_name
        elif isinstance(first_item, TableProxy):
            table_name = str(first_item)
        elif isinstance(first_item, str):
            first_item = first_item.split('.')
            table_name = f'{first_item[0]}.{first_item[1]}'
        res = res + f' FROM {table_name}'
        return res



    def replace_operators(self, where_clause:str):
        if not (self.db and self.db.py_db_operators):
            return where_clause
        for py_op, db_op in self.db.py_db_operators.items():
            where_clause = where_clause.replace(f' {py_op} ', f' {db_op} ')
        return where_clause



    def serialize_joins(self, joins:List[JoinOperation]):
        res = ''
        for join in joins:
            ser_join = self.serialize_join(join)
            res += ser_join        
        return res

    def serialize_join(self, join:JoinOperation):
        return f'{join.join_type} JOIN {join.table} on {join.on_clause} \n'

        

    def serialize_for_update(self, node: Savable, **kwargs) -> object:

        node_dict = self._serialize(node)

        if not ('node_id' in node_dict and node_dict['node_id']):
            raise DbError('unable update node without node_id')

        node_id = node_dict['node_id']
        pairs = ''

        for name, val in node_dict.items():
            pairs = pairs + name + ' = ' + val + ', '

        pairs = pairs[:-2]
        res = f'''UPDATE {self.full_name}
                SET {pairs}
                WHERE node_id = {node_id};'''
        return res


    def generate_id(self, resource: Savable = None):
        return uuid4()

    # def get_column_header_name(self, column_name:str):
    #     table_name = self.full_name
    #     res = f'{table_name}.{column_name}'
    #     alias = res.replace('.', '___')
    #     res = f'{res} as {alias}'
    #     return res


from .epure import epure as epure
# from .resource.gres.gres_db import GresDb as GresDb
# from .resource.db.constraint import Constraint, Default, Foreign, Prim, NotNull, Uniq, Check
# from .resource.db.constraint import Constraint as Constraint